# Author: Nathan Trouvain at 19/11/2021 <nathan.trouvain@inria.fr>
# Licence: MIT License
# Copyright: Xavier Hinaut (2018) <xavier.hinaut@inria.fr>
from collections import defaultdict
from typing import Iterable
from uuid import uuid4

import numpy as np

from .._base import check_xy
from .validation import is_mapping, is_sequence_set


def build_forward_sumodels(nodes, edges, already_trained):
    """Separate unfitted offline nodes from fitted nodes and gather all fitted
    nodes in submodels."""
    from ..model import Model

    offline_nodes = [
        n for n in nodes if n.is_trained_offline and n not in already_trained
    ]

    forward_nodes = list(set(nodes) - set(offline_nodes))
    forward_edges = [edge for edge in edges if edge[1] not in offline_nodes]

    submodel = Model(forward_nodes, forward_edges, name=f"SubModel-{uuid4()}")

    submodel.already_trained = already_trained

    return submodel, offline_nodes


def dist_states_to_next_subgraph(states, relations):
    """Map submodel output state vectors to input nodes of next submodel.

    Edges between first and second submodel are stored in 'relations'.
    """
    dist_states = {}
    for curr_node, next_nodes in relations.items():
        if len(next_nodes) > 1:
            for next_node in next_nodes:
                if dist_states.get(next_node) is None:
                    dist_states[next_node] = list()
                dist_states[next_node].append(states[curr_node])
        else:
            dist_states[next_nodes[0]] = states[curr_node]

    return dist_states


def allocate_returned_states(model, inputs, return_states=None):
    """Allocate output states matrices."""
    seq_len = inputs[list(inputs.keys())[0]].shape[0]

    # pre-allocate states
    if return_states == "all":
        states = {n.name: np.zeros((seq_len, n.output_dim)) for n in model.nodes}
    elif isinstance(return_states, Iterable):
        states = {
            n.name: np.zeros((seq_len, n.output_dim))
            for n in [model[name] for name in return_states]
        }
    else:
        states = {n.name: np.zeros((seq_len, n.output_dim)) for n in model.output_nodes}

    return states


def to_ragged_seq_set(data):
    """Convert dataset from mapping/array of sequences
    to lists of mappings of sequences."""
    # data is a dict
    if is_mapping(data):
        new_data = {}
        for name, datum in data.items():
            if not is_sequence_set(datum):
                # all sequences must at least be 2D (seq length, num features)
                # 1D sequences are converted to (1, num features) by default.
                new_datum = [np.atleast_2d(datum)]
            else:
                new_datum = datum
            new_data[name] = new_datum
        return new_data
    # data is an array or a list
    else:
        if not is_sequence_set(data):
            if data.ndim < 3:
                return [np.atleast_2d(data)]
            else:
                return data
        else:
            return data


def build_mapping(nodes, data, io_type="input"):
    """Map input/target data to input/trainable nodes in the model."""
    data = to_ragged_seq_set(data)
    if not is_mapping(data):
        if io_type == "input":
            data_map = {n.name: data for n in nodes}
        elif io_type == "target":
            # Remove unsupervised or already fitted nodes from the mapping
            data_map = {n.name: data for n in nodes if not n.unsupervised}
        else:
            raise ValueError(
                f"Unknown io_type: '{io_type}'. "
                f"Accepted io_types are 'input' and 'target'."
            )
    else:
        data_map = data.copy()

    return data_map


def unfold_mapping(data_map):
    """Convert a mapping of sequence lists into a list of sequence to nodes mappings."""
    seq_numbers = [len(data_map[n]) for n in data_map.keys()]
    if len(np.unique(seq_numbers)) > 1:
        seq_numbers = {n: len(data_map[n]) for n in data_map.keys()}
        raise ValueError(
            f"Found dataset with inconsistent number of sequences for each node. "
            f"Current number of sequences per node: {seq_numbers}"
        )

    # select an input dataset and check
    n_sequences = len(data_map[list(data_map.keys())[0]])

    mapped_sequences = []
    for i in range(n_sequences):
        sequence = {n: data_map[n][i] for n in data_map.keys()}
        mapped_sequences.append(sequence)

    return mapped_sequences


def fold_mapping(model, states, return_states):
    """Convert a list of sequence to nodes mappings into a mapping of lists or a
    simple array if possible."""
    n_sequences = len(states)
    if n_sequences == 1:
        states_map = states[0]
    else:
        states_map = defaultdict(list)
        for i in range(n_sequences):
            for node_name, seq in states[i].items():
                states_map[node_name] += [seq]

    if len(states_map) == 1 and return_states is None:
        return states_map[model.output_nodes[0].name]

    return states_map


def to_data_mapping(model, X, Y=None):
    """Map dataset to input/target nodes in the model."""
    X_map = build_mapping(model.input_nodes, X, io_type="input")

    Y_map = None
    if Y is not None:
        Y_map = build_mapping(model.trainable_nodes, Y, io_type="target")

    X_map, Y_map = check_xy(model, x=X_map, y=Y_map)

    X_sequences = unfold_mapping(X_map)

    if Y_map is None:
        n_sequences = len(X_sequences)
        Y_sequences = [None] * n_sequences
    else:
        Y_sequences = unfold_mapping(Y_map)

    return X_sequences, Y_sequences



# Author: Nathan Trouvain at 17/05/2022 <nathan.trouvain@inria.fr>
# Licence: MIT License
# Copyright: Xavier Hinaut (2018) <xavier.hinaut@inria.fr>
from functools import partial

import numpy as np

from ...mat_gen import zeros
from ...node import Node
from .base import (
    _assemble_wout,
    _compute_error,
    _initialize_readout,
    _prepare_inputs_for_learning,
    _split_and_save_wout,
    readout_forward,
)


def _rls(P, r, e, f):
    """Recursive Least Squares learning rule with forgetting factor."""
    f = float(f)    
    k = np.dot(P, r)
    rPr = np.dot(r.T, k).squeeze()
    c1 = float(1.0 / (f*(f + rPr)))
    c2 = float(1.0 / f)
    P = c2*P - c1 * np.outer(k, k)
    dw = -np.outer(e,np.dot(P, r))
    
    return dw, P


def train(node: "RLS", x, y=None):
    """Train a readout using RLS learning rule."""
    x, y = _prepare_inputs_for_learning(x, y, bias=node.input_bias, allow_reshape=True)

    error, r = _compute_error(node, x, y)

    P = node.P
    dw, P = _rls(P, r, error, node.forgetting)
    wo = _assemble_wout(node.Wout, node.bias, node.input_bias)
    wo = wo + dw.T

    _split_and_save_wout(node, wo)

    node.set_param("P", P)


def initialize(
    readout: "RLS", x=None, y=None, init_func=None, bias_init=None, bias=None
):

    _initialize_readout(readout, x, y, init_func, bias_init, bias)

    if x is not None:
        input_dim, alpha = readout.input_dim, readout.alpha

        if readout.input_bias:
            input_dim += 1

        P = np.eye(input_dim) / alpha

        readout.set_param("P", P)


class RLS(Node):
    """Single layer of neurons learning connections using Recursive Least Squares
    algorithm.

    The learning rules is well described in [1]_.
    The forgetting factor version of the RLS algorithm used here is described in [2]_.

    :py:attr:`RLS.params` **list**

    ================== =================================================================
    ``Wout``           Learned output weights (:math:`\\mathbf{W}_{out}`).
    ``bias``           Learned bias (:math:`\\mathbf{b}`).
    ``P``              Matrix :math:`\\mathbf{P}` of RLS rule.
    ================== =================================================================

    :py:attr:`RLS.hypers` **list**

    ================== =================================================================
    ``alpha``          Diagonal value of matrix P (:math:`\\alpha`) (:math:`1\\cdot 10^{-6}` by default).
    ``input_bias``     If True, learn a bias term (True by default).
    ``forgetting``     Forgetting factor (:math:`\\lambda`) (:math:`1` by default).  
    ================== =================================================================

    Parameters
    ----------
    output_dim : int, optional
        Number of units in the readout, can be inferred at first call.
    alpha : float or Python generator or iterable, default to 1e-6
        Diagonal value of matrix P.
    Wout : callable or array-like of shape (units, targets), default to :py:func:`~reservoirpy.mat_gen.zeros`
        Output weights matrix or initializer. If a callable (like a function) is
        used, then this function should accept any keywords
        parameters and at least two parameters that will be used to define the shape of
        the returned weight matrix.
    bias : callable or array-like of shape (units, 1), default to :py:func:`~reservoirpy.mat_gen.zeros`
        Bias weights vector or initializer. If a callable (like a function) is
        used, then this function should accept any keywords
        parameters and at least two parameters that will be used to define the shape of
        the returned weight matrix.
    input_bias : bool, default to True
        If True, then a bias parameter will be learned along with output weights.
    forgetting : float, default to 1.0
        The forgetting factor controls the weight given to past observations in the RLS update.
        A value less than 1.0 gives more weight to recent observations.
    name : str, optional
        Node name.

    References
    ----------

    .. [1] Sussillo, D., & Abbott, L. F. (2009). Generating Coherent Patterns of
           Activity from Chaotic Neural Networks. Neuron, 63(4), 544–557.
           https://doi.org/10.1016/j.neuron.2009.07.018

    .. [2] Waegeman, T., Wyffels, F., & Schrauwen, B. (2012). Feedback Control by Online
           Learning an Inverse Model. IEEE Transactions on Neural Networks and Learning
           Systems, 23(10), 1637–1648. https://doi.org/10.1109/TNNLS.2012.2208655

    Examples
    --------
    >>> x = np.random.normal(size=(100, 3))
    >>> noise = np.random.normal(scale=0.1, size=(100, 1))
    >>> y = x @ np.array([[10], [-0.2], [7.]]) + noise + 12.

    >>> from reservoirpy.nodes import RLS
    >>> rls_node = RLS(alpha=1e-1)

    >>> _ = rls_node.train(x[:5], y[:5])
    >>> print(rls_node.Wout.T, rls_node.bias)
    [[ 9.90731641 -0.06884784  6.87944632]] [[12.07802068]]
    >>> _ = rls_node.train(x[5:], y[5:])
    >>> print(rls_node.Wout.T, rls_node.bias)
    [[ 9.99223366 -0.20499636  6.98924066]] [[12.01128622]]
    """

    def __init__(
        self,
        output_dim=None,
        alpha=1e-6,
        Wout=zeros,
        bias=zeros,
        input_bias=True,
        forgetting=1.0,
        name=None,
    ):
        super(RLS, self).__init__(
            params={"Wout": None, "bias": None, "P": None},
            hypers={
                "alpha": alpha,
                "input_bias": input_bias,
                "forgetting": forgetting,
            },
            forward=readout_forward,
            train=train,
            initializer=partial(
                initialize, init_func=Wout, bias_init=bias, bias=input_bias
            ),
            output_dim=output_dim,
            name=name,
        )



# Author: Nathan Trouvain at 05/11/2021 <nathan.trouvain@inria.fr>
# Licence: MIT License
# Copyright: Xavier Hinaut (2018) <xavier.hinaut@inria.fr>
import numpy as np
import pytest
from numpy.testing import assert_array_almost_equal, assert_equal

from reservoirpy import set_seed
from reservoirpy.node import _filter_where_na_target
from reservoirpy.nodes import ESN, Reservoir, Ridge


def test_esn_init():
    esn = ESN(units=100, output_dim=1, lr=0.8, sr=0.4, ridge=1e-5, Win_bias=False)

    data = np.ones((1, 10))
    res = esn(data)

    assert esn.reservoir.W.shape == (100, 100)
    assert esn.reservoir.Win.shape == (100, 10)
    assert esn.reservoir.lr == 0.8
    assert esn.reservoir.units == 100

    data = np.ones((10000, 10))
    res = esn.run(data)

    assert res.shape == (10000, 1)

    with pytest.raises(ValueError):
        esn = ESN(units=100, output_dim=1, learning_method="foo")

    with pytest.raises(ValueError):
        esn = ESN(units=100, output_dim=1, reservoir_method="foo")


def test_esn_init_from_obj():
    res = Reservoir(100, lr=0.8, sr=0.4, input_bias=False)
    read = Ridge(1, ridge=1e-5)

    esn = ESN(reservoir=res, readout=read)

    data = np.ones((1, 10))
    res = esn(data)

    assert esn.reservoir.W.shape == (100, 100)
    assert esn.reservoir.Win.shape == (100, 10)
    assert esn.reservoir.lr == 0.8
    assert esn.reservoir.units == 100

    data = np.ones((10000, 10))
    res = esn.run(data)

    assert res.shape == (10000, 1)


def test_esn_states():
    res = Reservoir(100, lr=0.8, sr=0.4, input_bias=False)
    read = Ridge(1, ridge=1e-5)

    esn = ESN(reservoir=res, readout=read)

    data = np.ones((2, 10, 10))
    out = esn.run(data, return_states="all")

    assert out["reservoir"][0].shape == (10, 100)
    assert out["readout"][0].shape == (10, 1)

    out = esn.run(data, return_states=["reservoir"])

    assert out["reservoir"][0].shape == (10, 100)

    s_reservoir = esn.state()
    assert_equal(s_reservoir, res.state())

    s_readout = esn.state(which="readout")
    assert_equal(s_readout, read.state())

    with pytest.raises(ValueError):
        esn.state(which="foo")


def test_esn_feedback():
    esn = ESN(units=100, output_dim=5, lr=0.8, sr=0.4, ridge=1e-5, feedback=True)

    x = np.ones((1, 10))
    y = np.ones((1, 5))
    res = esn(x)

    assert esn.reservoir.W.shape == (100, 100)
    assert esn.reservoir.Win.shape == (100, 10)
    assert esn.readout.Wout.shape == (100, 5)
    assert res.shape == (1, 5)
    assert esn.reservoir.Wfb is not None
    assert esn.reservoir.Wfb.shape == (100, 5)

    esn.fit(x, y).run(x, forced_feedbacks=y)


def test_esn_parallel_fit_reproducibility():
    """Reproducibility of the ESN node across backends.
    Results may vary between OSes and NumPy versions.
    """
    seed = 1234
    rng = np.random.default_rng(seed=seed)
    X = list(rng.normal(0, 1, (10, 100, 10)))
    Y = [x @ rng.normal(0, 1, size=(10, 5)) for x in X]

    set_seed(seed)
    base_Wout = (
        ESN(
            units=100,
            ridge=1e-5,
            feedback=True,
            workers=-1,
            backend="sequential",
        )
        .fit(X, Y)
        .readout.Wout
    )

    for backend in ("loky", "multiprocessing", "threading", "sequential"):
        set_seed(seed)
        esn = ESN(
            units=100,
            ridge=1e-5,
            feedback=True,
            workers=-1,
            backend=backend,
        ).fit(X, Y)

        assert esn.reservoir.W.shape == (100, 100)
        assert esn.reservoir.Win.shape == (100, 10)
        assert esn.readout.Wout.shape == (100, 5)

        assert esn.reservoir.Wfb is not None
        assert esn.reservoir.Wfb.shape == (100, 5)
        assert np.abs(np.mean(esn.readout.Wout - base_Wout)) < 1e-14


def test_esn_parallel_run_reproducibility():
    """Reproducibility of the ESN node across backends. Results may
    vary between OSes and NumPy versions.
    """
    seed = 1000
    rng = np.random.default_rng(seed=seed)
    X = list(rng.normal(0, 1, (10, 100, 10)))
    Y = [x @ rng.normal(0, 1, size=(10, 5)) for x in X]

    set_seed(seed)
    # no feedback here. XXT and YXT sum orders are not deterministic
    # which results in small (float precision) differences across fits
    # and leads to error accumulation on run with feedback.
    set_seed(seed)
    esn = ESN(
        units=100,
        ridge=1e-5,
        workers=1,
        backend="sequential",
    ).fit(X, Y)

    base_y_out = esn.run(X[0])

    for backend in ("loky", "multiprocessing", "threading", "sequential"):
        set_seed(seed)
        esn = ESN(
            units=100,
            ridge=1e-5,
            workers=-1,
            backend=backend,
        ).fit(X, Y)

        y_out = esn.run(X[0])
        assert np.abs(np.mean(y_out - base_y_out)) < 1e-14


def test_hierarchical_esn_forbidden():
    esn1 = ESN(
        units=100,
        lr=0.8,
        sr=0.4,
        ridge=1e-5,
        feedback=True,
        workers=-1,
        backend="loky",
        name="E1",
    )

    esn2 = ESN(
        units=100,
        lr=0.8,
        sr=0.4,
        ridge=1e-5,
        feedback=True,
        workers=-1,
        backend="loky",
        name="E2",
    )

    # FrozenModel can't be linked (for now).
    with pytest.raises(TypeError):
        model = esn1 >> esn2


def test_na():

    # Define X, Y, Y_na
    X = np.ones((10, 5))
    Y = np.ones((10, 5))
    Y_na = Y
    Y_na[-1] = np.nan

    esn1 = ESN(
        units=100,
        lr=0.8,
        sr=0.4,
        ridge=1e-5,
        seed=1,
        workers=-1,
        feedback=True,
        backend="loky",
        name="E1",
    )

    esn2 = ESN(
        units=100,
        lr=0.8,
        sr=0.4,
        ridge=1e-5,
        seed=1,
        workers=-1,
        feedback=True,
        backend="loky",
        name="E2",
    )

    # Filtered the NaN data and get the predictions | Last row removed from X and Y
    X_filtered, Y_filtered = _filter_where_na_target(X, Y_na)
    esn1.fit(X_filtered, Y_filtered)
    pred_filtered = esn1.run(X_filtered)

    # Manually deleted the last row of the classic inputs
    esn2.fit(X[:-1, :], Y[:-1, :])
    pred_classic = esn2.run(X[:-1, :])

    # Assert that those predictions are equal
    assert_array_almost_equal(pred_filtered, pred_classic)



# Author: Nathan Trouvain at 12/07/2021 <nathan.trouvain@inria.fr>
# Licence: MIT License
# Copyright: Xavier Hinaut (2018) <xavier.hinaut@inria.fr>
from ..node import Node


def _io_initialize(io_node: "Node", x=None, **kwargs):
    if x is not None:
        if io_node.input_dim is None:
            io_node.set_input_dim(x.shape[1])
            io_node.set_output_dim(x.shape[1])


def _input_forward(inp_node: "Input", x):
    return x


class Input(Node):
    """Node feeding input data to other nodes in the models.

    Allow creating an input source and connecting it to several nodes at once.

    This node has no parameters and no hyperparameters.

    Parameters
    ----------
    input_dim : int
        Input dimension. Can be inferred at first call.
    name : str
        Node name.

    Example
    -------

    An input source feeding three different nodes in parallel.

    >>> from reservoirpy.nodes import Reservoir, Input
    >>> source = Input()
    >>> res1, res2, res3 = Reservoir(100), Reservoir(100), Reservoir(100)
    >>> model = source >> [res1, res2, res3]

    A model with different input sources. Use names to identify each source at runtime.

    >>> import numpy as np
    >>> from reservoirpy.nodes import Reservoir, Input
    >>> source1, source2 = Input(name="s1"), Input(name="s2")
    >>> res1, res2 = Reservoir(100), Reservoir(100)
    >>> model = source1 >> [res1, res2] & source2 >> [res1, res2]
    >>> outputs = model.run({"s1": np.ones((10, 5)), "s2": np.ones((10, 3))})
    """

    def __init__(self, input_dim=None, name=None, **kwargs):
        super(Input, self).__init__(
            forward=_input_forward,
            initializer=_io_initialize,
            input_dim=input_dim,
            output_dim=input_dim,
            name=name,
            **kwargs,
        )


class Output(Node):
    """Convenience node which can be used to add an output to a model.

    For instance, this node can be connected to a reservoir within a model to inspect
    its states.

    Parameters
    ----------
    name : str
        Node name.

    Example
    -------

    We can use the :py:class:`Output` node to probe the hidden states of Reservoir
    in an Echo State Network:

    >>> import numpy as np
    >>> from reservoirpy.nodes import Reservoir, Ridge, Output
    >>> reservoir = Reservoir(100)
    >>> readout = Ridge()
    >>> probe = Output(name="reservoir-states")
    >>> esn = reservoir >> readout & reservoir >> probe
    >>> _ = esn.initialize(np.ones((1,1)), np.ones((1,1)))

    When running the model, states can then be retrieved as an output:

    >>> data = np.ones((10, 1))
    >>> outputs = esn.run(data)
    >>> states = outputs["reservoir-states"]
    """

    def __init__(self, name=None, **kwargs):
        super(Output, self).__init__(
            forward=_input_forward, initializer=_io_initialize, name=name, **kwargs
        )



# Author: Nathan Trouvain at 10/11/2021 <nathan.trouvain@inria.fr>
# Licence: MIT License
# Copyright: Xavier Hinaut (2018) <xavier.hinaut@inria.fr>
import pytest

from .dummy_nodes import *


def test_node_link(plus_node, minus_node, offline_node, offline_node2, inverter_node):

    model1 = plus_node >> minus_node
    model2 = minus_node >> plus_node

    assert model1.edges == [(plus_node, minus_node)]
    assert model2.edges == [(minus_node, plus_node)]
    assert set(model1.nodes) == set(model2.nodes)

    model3 = plus_node >> offline_node
    model4 = minus_node >> offline_node2

    model = model3 >> model4

    assert set(model.edges) == {
        (plus_node, offline_node),
        (offline_node, minus_node),
        (minus_node, offline_node2),
    }
    assert set(model.nodes) == set(model3.nodes) | set(model4.nodes)

    # cycles in the model!
    with pytest.raises(RuntimeError):
        model1 & model2

    with pytest.raises(RuntimeError):
        plus_node >> minus_node >> plus_node

    with pytest.raises(RuntimeError):
        plus_node >> plus_node

    x = np.ones((1, 5))
    x2 = np.ones((1, 6))
    plus_node(x)
    minus_node(x2)

    # bad dimensions
    with pytest.raises(ValueError):
        plus_node >> minus_node

    with pytest.raises(ValueError):
        model1(x)

    # merge inplace on a node
    with pytest.raises(TypeError):
        plus_node &= minus_node


def test_node_link_several(plus_node, minus_node, offline_node):

    model = [plus_node, minus_node] >> offline_node

    assert len(model.nodes) == 4
    assert len(model.edges) == 3

    model = plus_node >> [offline_node, minus_node]

    assert set(model.nodes) == {plus_node, minus_node, offline_node}
    assert set(model.edges) == {(plus_node, offline_node), (plus_node, minus_node)}


def test_node_link_feedback(plus_node, minus_node):

    fb_plus_node = plus_node << minus_node

    assert id(fb_plus_node._feedback._sender) == id(minus_node)
    assert plus_node._feedback is None

    plus_node <<= minus_node
    assert id(plus_node._feedback._sender) == id(minus_node)


def test_model_merge(plus_node, minus_node, basic_offline_node):

    branch1 = plus_node >> minus_node
    branch2 = plus_node >> basic_offline_node

    model = branch1 & branch2

    assert set(model.nodes) == {plus_node, minus_node, basic_offline_node}
    assert set(model.edges) == {
        (plus_node, minus_node),
        (plus_node, basic_offline_node),
    }

    branch1 &= branch2

    assert set(branch1.nodes) == {plus_node, minus_node, basic_offline_node}
    assert set(branch1.edges) == {
        (plus_node, minus_node),
        (plus_node, basic_offline_node),
    }



# Author: Nathan Trouvain at 12/07/2021 <nathan.trouvain@inria.fr>
# Licence: MIT License
# Copyright: Xavier Hinaut (2018) <xavier.hinaut@inria.fr>
from collections import defaultdict, deque, namedtuple
from typing import Dict, List

import numpy as np

from .._base import _Node
from . import safe_defaultdict_copy
from .validation import is_mapping

DataPoint = namedtuple("DataPoint", "x, y")


def find_parents_and_children(edges):
    """Returns two dicts linking nodes to their parents and children in the graph."""
    parents = defaultdict(list)
    children = defaultdict(list)

    # Kludge to always have the parents and children in the same order at every run.
    # TODO: refactor the graphflow part.
    edges = sorted(list(edges), key=lambda x: x[0].name + x[1].name)

    for edge in edges:
        parent, child = edge
        parents[child] += [parent]
        children[parent] += [child]

    return parents, children


def topological_sort(nodes, edges, inputs=None):
    """Topological sort of nodes in a Model, to determine execution order."""
    if inputs is None:
        inputs, _ = find_entries_and_exits(nodes, edges)

    parents, children = find_parents_and_children(edges)

    # using Kahn's algorithm
    ordered_nodes = []
    edges = set(edges)
    inputs = deque(inputs)
    while len(inputs) > 0:
        n = inputs.pop()
        ordered_nodes.append(n)
        for m in children.get(n, ()):
            edges.remove((n, m))
            parents[m].remove(n)
            if parents.get(m) is None or len(parents[m]) < 1:
                inputs.append(m)
    if len(edges) > 0:
        raise RuntimeError(
            "Model has a cycle: impossible "
            "to automatically determine operations "
            "order in the model."
        )
    else:
        return ordered_nodes


def get_offline_subgraphs(nodes, edges):
    """Cut a graph into several subgraphs where output nodes are untrained offline
    learner nodes."""
    inputs, outputs = find_entries_and_exits(nodes, edges)
    parents, children = find_parents_and_children(edges)

    offlines = set(
        [n for n in nodes if n.is_trained_offline and not n.is_trained_online]
    )
    included, trained = set(), set()
    subgraphs, required = [], []
    _nodes = nodes.copy()
    while trained != offlines:
        subnodes, subedges = [], []
        for node in _nodes:
            if node in inputs or all([p in included for p in parents.get(node)]):

                if node.is_trained_offline and node not in trained:
                    trained.add(node)
                    subnodes.append(node)
                else:
                    if node not in outputs:
                        subnodes.append(node)
                    included.add(node)

        subedges = [
            edge for edge in edges if edge[0] in subnodes and edge[1] in subnodes
        ]
        subgraphs.append((subnodes, subedges))
        _nodes = [n for n in nodes if n not in included]

    required = _get_required_nodes(subgraphs, children)

    return list(zip(subgraphs, required))


def _get_required_nodes(subgraphs, children):
    """Get nodes whose outputs are required to run/fit children nodes."""
    req = []
    fitted = set()
    for i in range(1, len(subgraphs)):
        currs = set(subgraphs[i - 1][0])
        nexts = set(subgraphs[i][0])

        req.append(_get_links(currs, nexts, children))

        fitted |= set([node for node in currs if node.is_trained_offline])

    nexts = set(
        [n for n in subgraphs[-1][0] if n.is_trained_offline and n not in fitted]
    )
    currs = set(
        [n for n in subgraphs[-1][0] if not n.is_trained_offline or n in fitted]
    )

    req.append(_get_links(currs, nexts, children))

    return req


def _get_links(previous, nexts, children):
    """Returns graphs edges between two subgraphs."""
    links = {}
    for n in previous:
        next_children = []
        if n not in nexts:
            next_children = [c.name for c in children.get(n, []) if c in nexts]

        if len(next_children) > 0:
            links[n.name] = next_children

    return links


def find_entries_and_exits(nodes, edges):
    """Find outputs and inputs nodes of a directed acyclic graph."""
    nodes = set(nodes)
    senders = set([n for n, _ in edges])
    receivers = set([n for _, n in edges])

    lonely = nodes - senders - receivers

    entrypoints = senders - receivers | lonely
    endpoints = receivers - senders | lonely

    return list(entrypoints), list(endpoints)


def dispatch(
    X,
    Y=None,
    shift_fb=True,
    return_targets=False,
    force_teachers=True,
):
    """Transform data from a dict of arrays
    ([node], timesteps, dimension) to an iterator yielding
    a node: data mapping for each timestep."""
    X_map, Y_map = X, Y
    current_node = list(X_map.keys())[0]
    sequence_length = len(X_map[current_node])

    for i in range(sequence_length):
        x = {node: X_map[node][np.newaxis, i] for node in X_map.keys()}
        if Y_map is not None:
            y = None
            if return_targets:
                y = {node: Y_map[node][np.newaxis, i] for node in Y_map.keys()}
            # if feedbacks vectors are meant to be fed
            # with a delay in time of one timestep w.r.t. 'X_map'
            if shift_fb:
                if i == 0:
                    if force_teachers:
                        fb = {
                            node: np.zeros_like(Y_map[node][np.newaxis, i])
                            for node in Y_map.keys()
                        }
                    else:
                        fb = {node: None for node in Y_map.keys()}
                else:
                    fb = {node: Y_map[node][np.newaxis, i - 1] for node in Y_map.keys()}
            # else assume that all feedback vectors must be instantaneously
            # fed to the network. This means that 'Y_map' already contains
            # data that is delayed by one timestep w.r.t. 'X_map'.
            else:
                fb = {node: Y_map[node][np.newaxis, i] for node in Y_map.keys()}
        else:
            fb = y = None

        yield x, fb, y


class DataDispatcher:
    """A utility used to feed data to nodes in a Model."""

    _inputs: List
    _parents: Dict

    def __init__(self, model):
        self._nodes = model.nodes
        self._trainables = model.trainable_nodes
        self._inputs = model.input_nodes
        self.__parents, _ = find_parents_and_children(model.edges)

        self._parents = safe_defaultdict_copy(self.__parents)
        self._teachers = dict()

    def __getitem__(self, item):
        return self.get(item)

    def _check_inputs(self, input_mapping):
        if is_mapping(input_mapping):
            for node in self._inputs:
                if input_mapping.get(node.name) is None:
                    raise KeyError(
                        f"Node {node.name} not found "
                        f"in data mapping. This node requires "
                        f"data to run."
                    )

    def _check_targets(self, target_mapping):
        if is_mapping(target_mapping):
            for node in self._nodes:
                if (
                    node in self._trainables
                    and not node.fitted
                    and target_mapping.get(node.name) is None
                ):
                    raise KeyError(
                        f"Trainable node {node.name} not found "
                        f"in target/feedback data mapping. This "
                        f"node requires "
                        f"target values."
                    )

    def get(self, item):
        parents = self._parents.get(item, ())
        teacher = self._teachers.get(item, None)

        x = []
        for parent in parents:
            if isinstance(parent, _Node):
                x.append(parent.state())
            else:
                x.append(parent)

        # in theory, only operators can support several incoming signal
        # i.e. several operands, so unpack data if the list is unecessary
        if len(x) == 1:
            x = x[0]

        return DataPoint(x=x, y=teacher)

    def load(self, X=None, Y=None):
        """Load input and target data for dispatch."""
        self._parents = safe_defaultdict_copy(self.__parents)
        self._teachers = dict()

        if X is not None:
            self._check_inputs(X)
            if is_mapping(X):
                for node in self._nodes:
                    if X.get(node.name) is not None:
                        self._parents[node] += [X[node.name]]

            else:
                for inp_node in self._inputs:
                    self._parents[inp_node] += [X]

        if Y is not None:
            self._check_targets(Y)
            for node in self._nodes:
                if is_mapping(Y):
                    if Y.get(node.name) is not None:
                        self._teachers[node] = Y.get(node.name)
                else:
                    if node in self._trainables:
                        self._teachers[node] = Y
        return self



# @author: Xavier HINAUT
# xavier.hinaut@inria.fr
# Copyright Xavier Hinaut 2018
# We would like to thank Mantas Lukosevicius for his code that
# was used as inspiration for this code:
# # http://minds.jacobs-university.de/mantas/code
import time
from typing import Callable, Sequence

import numpy as np

from ..type import Activation, Data
from ._base import _ESNBase
from .regression_models import RidgeRegression
from .utils.parallel import parallelize
from .utils.validation import _check_values, check_input_lists


def _get_offline_model(ridge: float = 0.0, dtype: np.dtype = np.float64):
    return RidgeRegression(ridge, dtype=dtype)


class ESN(_ESNBase):
    """Base class of Echo State Networks.

    Simple, fast, parallelizable and object-oriented
    implementation of Echo State Networks [#]_ [#]_, using offline
    learning methods.

    Warning
    -------

        The v0.2 model :py:class:`compat.ESN` is deprecated. Consider using
        the new Node API introduced in v0.3 (see :ref:`node`).

    The :py:class:`compat.ESN` class is the angular stone of ReservoirPy
    offline learning methods using reservoir computing.
    Echo State Network allows one to:

    - quickly build ESNs, using the :py:mod:`reservoirpy.mat_gen`
      module to initialize weights,
    - train and test ESNs on the task of your choice,
    - use the trained ESNs on the task of your choice,
      either in predictive mode or generative mode.

    Parameters
    ----------
        lr: float
            Leaking rate
        W: np.ndarray
            Reservoir weights matrix
        Win: np.ndarray
            Input weights matrix
        input_bias: bool, optional
            If True, will add a constant bias
            to the input vector. By default, True.
        reg_model: Callable, optional
            A scikit-learn linear model function to use for
            regression. Should be None if ridge is used.
        ridge: float, optional
            Ridge regularization coefficient for Tikonov regression.
            Should be None if reg_model is used. By default, pseudo-inversion
            of internal states and teacher signals is used.
        Wfb: np.array, optional
            Feedback weights matrix.
        fbfunc: Callable, optional
            Feedback activation function.
        typefloat: numpy.dtype, optional

    Attributes
    ----------
        lr : float
            Leaking rate.
        activation: Callable
            Reservoir activation function.
        fbfunc: Callable
            Feedback activation function.
        noise_in: float
            Input noise gain.
        noise_rc: float
            Reservoir states noise gain.
        noise_out: float
            Feedback noise gain.
        seed: int
            Random state seed.
        typefloat: numpy.dtype

    References
    ----------

        .. [#] H. Jaeger, ‘The “echo state” approach to analysing
               and training recurrent neural networks – with an
               Erratum note’, p. 48.
        .. [#] M. Lukoševičius, ‘A Practical Guide to Applying Echo
               State Networks’, Jan. 2012, doi: 10.1007/978-3-642-35289-8_36.
    """

    def __init__(
        self,
        lr: float,
        W: np.ndarray,
        Win: np.ndarray,
        input_bias: bool = True,
        reg_model: Callable = None,
        ridge: float = 0.0,
        Wfb: np.ndarray = None,
        fbfunc: Callable = lambda x: x,
        noise_in: float = 0.0,
        noise_rc: float = 0.0,
        noise_out: float = 0.0,
        activation: Activation = np.tanh,
        seed: int = None,
        typefloat: np.dtype = np.float64,
    ):
        super(ESN, self).__init__(
            W,
            Win,
            lr=lr,
            input_bias=input_bias,
            activation=activation,
            Wfb=Wfb,
            fbfunc=fbfunc,
            Wout=None,
            noise_in=noise_in,
            noise_rc=noise_rc,
            noise_out=noise_out,
            seed=seed,
            typefloat=typefloat,
        )
        self.model = _get_offline_model(ridge, dtype=typefloat)

    @property
    def ridge(self):
        """L2 regularization coefficient for readout fitting."""
        return getattr(self.model, "ridge", None)

    @ridge.setter
    def ridge(self, value):
        if hasattr(self.model, "ridge"):
            self.model.ridge = value

    def fit_readout(
        self,
        states: Data,
        teachers: Data,
        reg_model: Callable = None,
        ridge: float = None,
        force_pinv: bool = False,
        verbose: bool = False,
    ) -> np.ndarray:
        """Compute a readout matrix by fitting the states computed by the ESN
        to the expected values, using the regression model defined
        in the ESN.

        Parameters
        ----------
            states: list of numpy.ndarray
                All states computed.

            teachers: list of numpy.ndarray
                All ground truth vectors.

            reg_model: scikit-learn regression model, optional
                A scikit-learn regression model to use for readout
                weights computation.

            ridge: float, optional
                Use Tikhonov regression for readout weights computation
                and set regularization parameter to the parameter value.

            force_pinv: bool, optional
                Overwrite all previous parameters and
                force computation of readout using pseudo-inversion.

            verbose: bool, optional

        Returns
        -------
            numpy.ndarray
                Readout matrix.
        """
        states, teachers = check_input_lists(states, self.N, teachers, self.dim_out)

        # switch the regression model used at instantiation if needed.
        # WARNING: this change won't be saved by the save function.
        if ridge is not None:
            offline_model = _get_offline_model(ridge, dtype=self.typefloat)
        elif force_pinv:
            offline_model = _get_offline_model(ridge=0.0)
        else:
            offline_model = self.model

        # check if network responses are valid
        _check_values(array_or_list=states, value=None)

        if verbose:
            tic = time.time()
            print("Linear regression...")

        self.Wout = offline_model.fit(X=states, Y=teachers)

        if verbose:
            toc = time.time()
            print(f"Linear regression done! (in {toc - tic} sec)")

        return self.Wout

    def train(
        self,
        inputs: Data,
        teachers: Data,
        wash_nr_time_step: int = 0,
        workers: int = -1,
        seed: int = None,
        verbose: bool = False,
        backend=None,
        use_memmap=None,
        return_states: bool = False,
    ) -> Sequence[np.ndarray]:
        """Train the ESN model on set of input sequences.

        Parameters
        ----------
            inputs: list of numpy.ndarray
                List of inputs.
                Note that it should always be a list of sequences, i.e. if
                only one sequence (array with rows representing time axis)
                of inputs is used, it should be alone in a list.
            teachers: list of numpy.ndarray
                List of ground truths.
                Note that is should always be a list of
                sequences of the same length than the `inputs`, i.e. if
                only one sequence of inputs is used, it should be alone in a
                list.
            wash_nr_time_step: int
                Number of states to considered as transient when training. Transient
                states will be discarded when computing readout matrix. By default,
                no states are removes.
            workers: int, optional
                If n >= 1, will enable parallelization of states computation with
                n threads/processes, if possible. If n = -1, will use all available
                resources for parallelization. By default, -1.
            return_states: bool, False by default
                If `True`, the function will return all the internal states computed
                during the training. Be warned that this may be too heavy for the
                memory of your computer.
            backend:
                kept for compatibility with previous versions.
            use_memmap:
                kept for compatibility with previous versions.
            verbose: bool, optional

        Returns
        -------
            list of numpy.ndarray
                All states computed, for all inputs.

        Note
        ----
            If only one input sequence is provided ("continuous time" inputs),
            workers should be 1, because parallelization is impossible. In other
            cases, if using large NumPy arrays during computation (which is often
            the case), prefer using `threading` backend to avoid huge overhead.
            Multiprocess is a good idea only in very specific cases, and this code
            is not (yet) well suited for this.
        """
        # autochecks of inputs and outputs
        inputs, teachers = check_input_lists(
            inputs, self.dim_in, teachers, self.dim_out
        )

        self._dim_out = teachers[0].shape[1]
        self.model.initialize(self.N, self.dim_out)

        lengths = [i.shape[0] for i in inputs]
        steps = sum(lengths)

        if verbose:
            print(
                f"Training on {len(inputs)} inputs ({steps} steps) "
                f"-- wash: {wash_nr_time_step} steps"
            )

        def train_fn(*, x, y, pbar):
            s = self._compute_states(x, y, seed=seed, pbar=pbar)
            self.model.partial_fit(
                s[wash_nr_time_step:], y
            )  # increment X.X^T and Y.X^T
            # or save data for sklearn fit
            return s

        _, states = parallelize(
            self,
            train_fn,
            workers,
            lengths,
            return_states,
            pbar_text="Train",
            verbose=verbose,
            x=inputs,
            y=teachers,
        )

        self.Wout = self.model.fit()  # perform Y.X^T.(X.X^T + ridge)^-1
        # or sklearn fit

        return states



# Author: Nathan Trouvain at 23/06/2021 <nathan.trouvain@inria.fr>
# Licence: MIT License
# Copyright: Xavier Hinaut (2018) <xavier.hinaut@inria.fr>
import numpy as np
import pytest
from numpy.testing import assert_array_equal
from scipy.sparse import csr_matrix

from ..utils.validation import check_input_lists, check_reservoir_matrices


@pytest.mark.parametrize(
    "X,Y",
    (
        ([np.ones((10, 5)) for _ in range(10)], [np.ones((10, 2)) for _ in range(9)]),
        (
            [np.ones((10, 5)) for _ in range(9)] + [np.ones((9, 5))],
            [np.ones((10, 2)) for _ in range(10)],
        ),
        (
            [np.ones((10, 5)) for _ in range(10)],
            [np.ones((10, 3))] + [np.ones((10, 2)) for _ in range(9)],
        ),
    ),
)
def test_bad_input_list(X, Y):
    with pytest.raises(ValueError):
        check_input_lists(X, dim_in=5, Y=Y, dim_out=2)


def test_good_input_list():
    X = [np.ones((10, 5)) for _ in range(10)]
    Y = [np.ones((10, 2)) for _ in range(10)]

    X1, Y1 = check_input_lists(X, dim_in=5, Y=Y, dim_out=2)

    assert all([np.all(x0 == x1) for x0, x1 in zip(X, X1)])
    assert all([np.all(x0 == x1) for x0, x1 in zip(Y, Y1)])


def test_bad_matrices_data():
    W = np.ones((10, 10))
    W[1, 1] = np.nan
    Win = np.ones((2, 10))

    with pytest.raises(ValueError):
        check_reservoir_matrices(W, Win)

    W = np.ones((10, 10))
    Win = np.ones((2, 10))
    Win[0, 3] = np.inf

    with pytest.raises(ValueError):
        check_reservoir_matrices(W, Win)

    W = np.ones((10, 10))
    Win = np.ones((2, 10))
    Wout = np.ones((2, 10))
    Wout[0, 0] = None

    with pytest.raises(ValueError):
        check_reservoir_matrices(W, Win, Wout=Wout)


def test_bad_matrices_type():
    W = np.ones((10, 10))
    Win = np.ones((2, 10)).astype(str)

    with pytest.raises(TypeError):
        check_reservoir_matrices(W, Win)

    W = np.ones((10, 10)).tolist()
    W[0][0] = "a"
    Win = np.ones((2, 10))

    with pytest.raises(TypeError):
        check_reservoir_matrices(W, Win)


def test_bad_matrices_shapes():
    W = np.ones((5, 10))
    Win = np.ones((2, 10))

    with pytest.raises(ValueError):
        check_reservoir_matrices(W, Win)

    W = np.ones((10, 10))
    Win = np.ones((9, 2))

    with pytest.raises(ValueError):
        check_reservoir_matrices(W, Win)

    W = np.ones((10, 10))
    Win = np.ones((10, 2))
    Wout = np.zeros((10, 1))

    with pytest.raises(ValueError):
        check_reservoir_matrices(W, Win, Wout=Wout)

    W = np.ones((10, 10))
    Win = np.ones((10, 2))
    Wout = np.zeros((11, 1))
    Wfb = np.ones((10, 2))

    with pytest.raises(ValueError):
        check_reservoir_matrices(W, Win, Wout=Wout, Wfb=Wfb)


def test_good_matrices():
    W = np.ones((10, 10))
    Win = np.ones((10, 2))
    Wout = np.zeros((1, 11))
    Wfb = np.ones((10, 1))

    W1, Win1, Wout1, Wfb1 = check_reservoir_matrices(W, Win, Wout=Wout, Wfb=Wfb)

    assert_array_equal(W1, W)
    assert_array_equal(Win1, Win)
    assert_array_equal(Wout1, Wout)
    assert_array_equal(Wfb1, Wfb)


def test_good_sparse_matrices():
    W = csr_matrix(np.ones((10, 10)))
    Win = csr_matrix(np.ones((10, 2)))
    Wout = csr_matrix(np.zeros((1, 11)))
    Wfb = csr_matrix(np.ones((10, 1)))

    W1, Win1, Wout1, Wfb1 = check_reservoir_matrices(W, Win, Wout=Wout, Wfb=Wfb)

    assert_array_equal(W1.toarray(), W.toarray())
    assert_array_equal(Win1.toarray(), Win.toarray())
    assert_array_equal(Wout1.toarray(), Wout.toarray())
    assert_array_equal(Wfb1.toarray(), Wfb.toarray())



# Author: Nathan Trouvain at 17/05/2022 <nathan.trouvain@inria.fr>
# Licence: MIT License
# Copyright: Xavier Hinaut (2018) <xavier.hinaut@inria.fr>
from functools import partial
from numbers import Number
from typing import Iterable

import numpy as np

from ...mat_gen import zeros
from ...node import Node
from .base import (
    _assemble_wout,
    _compute_error,
    _initialize_readout,
    _prepare_inputs_for_learning,
    _split_and_save_wout,
    readout_forward,
)


def _lms(alpha, r, e):
    """Least Mean Squares learning rule."""
    # learning rate is a generator to allow scheduling
    dw = -next(alpha) * np.outer(e, r)
    return dw


def train(node: "LMS", x, y=None):
    """Train a readout using LMS learning rule."""
    x, y = _prepare_inputs_for_learning(x, y, bias=node.input_bias, allow_reshape=True)

    error, r = _compute_error(node, x, y)

    alpha = node._alpha_gen
    dw = _lms(alpha, r, error)
    wo = _assemble_wout(node.Wout, node.bias, node.input_bias)
    wo = wo + dw.T

    _split_and_save_wout(node, wo)


def initialize(
    readout: "LMS", x=None, y=None, init_func=None, bias_init=None, bias=None
):
    _initialize_readout(readout, x, y, init_func, bias_init, bias)


class LMS(Node):
    """Single layer of neurons learning connections using Least Mean Squares
    algorithm.

    The learning rules is well described in [1]_.

    :py:attr:`LMS.params` **list**

    ================== =================================================================
    ``Wout``           Learned output weights (:math:`\\mathbf{W}_{out}`).
    ``bias``           Learned bias (:math:`\\mathbf{b}`).
    ``P``              Matrix :math:`\\mathbf{P}` of RLS rule.
    ================== =================================================================

    :py:attr:`LMS.hypers` **list**

    ================== =================================================================
    ``alpha``          Learning rate (:math:`\\alpha`) (:math:`1\\cdot 10^{-6}` by default).
    ``input_bias``     If True, learn a bias term (True by default).
    ================== =================================================================

    Parameters
    ----------
    output_dim : int, optional
        Number of units in the readout, can be inferred at first call.
    alpha : float or Python generator or iterable, default to 1e-6
        Learning rate. If an iterable or a generator is provided, the learning rate can
        be changed at each timestep of training. A new learning rate will be drawn from
        the iterable or generator at each timestep.
    Wout : callable or array-like of shape (units, targets), default to :py:func:`~reservoirpy.mat_gen.zeros`
        Output weights matrix or initializer. If a callable (like a function) is
        used, then this function should accept any keywords
        parameters and at least two parameters that will be used to define the shape of
        the returned weight matrix.
    bias : callable or array-like of shape (units, 1), default to :py:func:`~reservoirpy.mat_gen.zeros`
        Bias weights vector or initializer. If a callable (like a function) is
        used, then this function should accept any keywords
        parameters and at least two parameters that will be used to define the shape of
        the returned weight matrix.
    input_bias : bool, default to True
        If True, then a bias parameter will be learned along with output weights.
    name : str, optional
        Node name.

    Examples
    --------
    >>> x = np.random.normal(size=(100, 3))
    >>> noise = np.random.normal(scale=0.01, size=(100, 1))
    >>> y = x @ np.array([[10], [-0.2], [7.]]) + noise + 12.

    >>> from reservoirpy.nodes import LMS
    >>> lms_node = LMS(alpha=1e-1)

    >>> lms_node.train(x[:50], y[:50])
    >>> print(lms_node.Wout.T, lms_node.bias)
    [[ 9.156 -0.967   6.411]] [[11.564]]
    >>> lms_node.train(x[50:], y[50:])
    >>> print(lms_node.Wout.T, lms_node.bias)
    [[ 9.998 -0.202  7.001]] [[12.005]]

    References
    ----------

    .. [1] Sussillo, D., & Abbott, L. F. (2009). Generating Coherent Patterns of
           Activity from Chaotic Neural Networks. Neuron, 63(4), 544–557.
           https://doi.org/10.1016/j.neuron.2009.07.018
    """

    def __init__(
        self,
        output_dim=None,
        alpha=1e-6,
        Wout=zeros,
        bias=zeros,
        input_bias=True,
        name=None,
    ):
        if isinstance(alpha, Number):

            def _alpha_gen():
                while True:
                    yield alpha

            alpha_gen = _alpha_gen()
        elif isinstance(alpha, Iterable):
            alpha_gen = alpha
        else:
            raise TypeError(
                "'alpha' parameter should be a float or an iterable yielding floats."
            )

        super(LMS, self).__init__(
            params={"Wout": None, "bias": None},
            hypers={
                "alpha": alpha,
                "_alpha_gen": alpha_gen,
                "input_bias": input_bias,
            },
            forward=readout_forward,
            train=train,
            initializer=partial(
                initialize, init_func=Wout, bias_init=bias, bias=input_bias
            ),
            output_dim=output_dim,
            name=name,
        )



import numpy as np

from ..delay import Delay
from ..readouts import Ridge


def test_initialize():
    delay1 = Delay(delay=10)
    delay1.run(np.ones((10, 2)))
    assert delay1.input_dim == 2
    assert np.all(delay1.buffer[0] == 1.0)

    delay2 = Delay(delay=10, input_dim=5)
    delay2.initialize()
    assert delay2.input_dim == 5
    assert np.all(delay2.buffer[0] == 0.0)

    delay3 = Delay(delay=10, initial_values=np.ones((10, 7)))
    delay3.initialize()
    assert delay3.input_dim == 7
    assert np.all(delay3.buffer[0] == 1.0)


def test_no_delay():
    delay_node = Delay(delay=0)

    x = np.array([0.2, 0.3])
    y = delay_node(x)
    assert np.all(x == y)

    x = np.linspace(1, 12, num=12).reshape(-1, 2)
    y = delay_node.run(x)
    assert np.all(x == y)


def test_1_delay():
    delay_node = Delay(delay=1)

    x1 = np.array([0.2, 0.3])
    y = delay_node(x1)
    assert np.all(y == 0.0)

    x2 = np.linspace(1, 12, num=12).reshape(-1, 2)
    y = delay_node.run(x2)
    assert np.all(y[0] == x1)
    assert np.all(y[1:] == x2[:-1])


def test_large_delay():
    # Note: this is quite slow... is deque the best format?
    delay_node = Delay(delay=1_000)

    x = np.array([0.2, 0.3])
    y = delay_node(x)
    assert np.all(y == 0.0)
    assert np.all(delay_node.buffer[0] == x)
    assert np.all(delay_node.buffer[-1] == 0.0)

    delay_node.run(np.zeros((999, 2)))
    y = delay_node(np.zeros((1, 2)))
    assert np.all(y == x)


def test_multiseries_delay():
    delay_node = Delay(delay=2)
    readout = Ridge(ridge=1e-3)
    model = delay_node >> readout

    x = list(np.fromfunction(lambda i, j, k: i + j, (2, 4, 2)))
    y = list(np.fromfunction(lambda i, j, k: i + j, (2, 4, 1)))

    model.fit(x, y, warmup=2)



import time
import warnings
from abc import ABCMeta
from functools import partial
from pathlib import Path
from typing import Sequence, Tuple, Union

import numpy as np
from joblib import Parallel, delayed
from numpy.random import Generator, SeedSequence, default_rng
from tqdm import tqdm

from ..type import Activation, Data, Weights
from ..utils.parallel import get_joblib_backend
from ..utils.validation import add_bias
from .utils.parallel import ParallelProgressQueue, parallelize
from .utils.save import _save
from .utils.validation import check_input_lists, check_reservoir_matrices


class _ESNBase(metaclass=ABCMeta):
    def __init__(
        self,
        W: Weights,
        Win: Weights,
        lr: float = 1.0,
        input_bias: bool = True,
        activation: Activation = np.tanh,
        Wfb: Weights = None,
        fbfunc: Activation = lambda x: x,
        Wout: Weights = None,
        noise_in: float = 0.0,
        noise_rc: float = 0.0,
        noise_out: float = 0.0,
        seed: int = None,
        typefloat: np.dtype = np.float64,
    ):

        W, Win, Wout, Wfb = check_reservoir_matrices(
            W, Win, Wout=Wout, Wfb=Wfb, caller=self
        )

        self._W = W
        self._Win = Win
        self._Wout = Wout
        self._Wfb = Wfb

        self.lr = lr
        self.activation = activation
        self.noise_in = noise_in
        self.noise_rc = noise_rc
        self.noise_out = noise_out
        self.typefloat = typefloat
        self.seed = seed
        self.fbfunc = fbfunc
        self._input_bias = input_bias

        self._N = self._W.shape[0]
        self._dim_in = self._Win.shape[1]
        if self._input_bias:
            self._dim_in = self._dim_in - 1

        if self._Wout is not None:
            self._dim_out = self._Wout.shape[0]
        elif self._Wfb is not None:
            self._dim_out = self._Wfb.shape[1]
        else:
            self._dim_out = None

    def __repr__(self):
        trained = True
        if self.Wout is None:
            trained = False
        fb = self.Wfb is not None
        name = self.__class__.__name__
        out = (
            f"{name}(trained={trained}, feedback={fb}, N={self.N}, "
            f"lr={self.lr}, input_bias={self.input_bias}, input_dim={self.dim_in})"
        )
        return out

    @property
    def N(self):
        """Number of units."""
        return self._N

    @property
    def dim_in(self):
        """Input dimension."""
        return self._dim_in

    @property
    def dim_out(self):
        """Output (readout) dimension."""
        return self._dim_out

    @property
    def input_bias(self):
        """If True, constant bias is added to inputs."""
        return self._input_bias

    @property
    def use_raw_input(self):
        """If True, raw inputs are concatenated to states before readout."""
        return self._use_raw_input

    @property
    def Win(self):
        """Input weight matrix."""
        return self._Win

    @Win.setter
    def Win(self, matrix):
        _, Win, _, _ = check_reservoir_matrices(
            self._W, matrix, self._Wout, self._Wfb, caller=self
        )
        self._Win = Win
        self._dim_in = self._Win.shape[1]
        if self._input_bias:
            self._dim_in = self._dim_in - 1

    @property
    def W(self):
        """Recurrent weight matrix."""
        return self._W

    @W.setter
    def W(self, matrix):
        W, _, _, _ = check_reservoir_matrices(
            matrix, self._Win, self._Wout, self._Wfb, caller=self
        )
        self._W = W
        self._N = self._W.shape[0]

    @property
    def Wfb(self):
        """Feedback weight matrix."""
        return self._Wfb

    @Wfb.setter
    def Wfb(self, matrix):
        _, _, _, Wfb = check_reservoir_matrices(
            self._W, self._Win, self._Wout, matrix, caller=self
        )
        self._Wfb = Wfb
        self._dim_out = self._Wfb.shape[1]

    @property
    def Wout(self):
        """Readout weight matrix."""
        return self._Wout

    @Wout.setter
    def Wout(self, matrix):
        _, _, Wout, _ = check_reservoir_matrices(
            self._W, self._Win, matrix, self._Wfb, caller=self
        )
        self._Wout = Wout
        self._dim_out = self._Wout.shape[0]

    def zero_state(self):
        """Returns zero state vector."""
        return np.zeros((1, self.N), dtype=self.typefloat)

    def zero_feedback(self):
        """Returns a zero feedback vector."""
        if self.Wfb is not None:
            return np.zeros((1, self.dim_out), dtype=self.typefloat)
        else:
            return None

    def _get_next_state(
        self,
        single_input: np.ndarray,
        feedback: np.ndarray = None,
        last_state: np.ndarray = None,
        noise_generator: Generator = None,
    ) -> np.ndarray:
        """Given a state vector x(t) and an input vector u(t),
        compute the state vector x(t+1).

        Parameters
        ----------
            single_input: np.ndarray
                Input vector u(t)

            feedback: numpy.ndarray, optional
                Feedback vector if enabled.
            last_state: numpy.ndarray, optional
                Current state to update x(t). Default to 0 vector.

        Raises
        ------
            RuntimeError: feedback is enabled but no feedback vector is available.

        Returns
        -------
            numpy.ndarray
                Next state x(t+1)
        """
        if noise_generator is None:
            noise_generator = default_rng()

        # first initialize the current state of the ESN
        if last_state is None:
            x = self.zero_state()
        else:
            x = np.asarray(last_state, dtype=self.typefloat).reshape(1, -1)

        u = np.asarray(single_input, dtype=self.typefloat).reshape(1, -1)
        # add bias
        if self._input_bias:
            u = add_bias(u)

        # prepare noise sequence
        noise_in = self.noise_in * noise_generator.uniform(-1, 1, size=u.shape)
        noise_rc = self.noise_rc * noise_generator.uniform(-1, 1, size=x.shape)

        # linear transformation
        x1 = (u + noise_in) @ self.Win.T + x @ self.W

        # add feedback if requested
        if self.Wfb is not None:
            if feedback is None:
                feedback = self.zero_feedback()

            noise_out = self.noise_out * noise_generator.uniform(
                -1, 1, size=feedback.shape
            )
            fb = self.fbfunc(np.asarray(feedback)).reshape(1, -1)
            x1 += (fb + noise_out) @ self.Wfb.T

        # previous states memory leak and non-linear transformation
        x1 = (1 - self.lr) * x + self.lr * (self.activation(x1) + noise_rc)

        # return the next state computed
        return x1

    def _compute_states(
        self,
        input: np.ndarray,
        forced_teacher: np.ndarray = None,
        init_state: np.ndarray = None,
        init_fb: np.ndarray = None,
        seed: int = None,
        verbose: bool = False,
        **kwargs,
    ) -> Union[Tuple[np.ndarray, np.ndarray], np.ndarray]:
        """Compute all states generated from a single sequence of inputs.

        Parameters
            input {np.ndarray} -- Sequence of inputs.

        Keyword Arguments:
            forced_teacher {np.ndarray} -- Ground truth vectors to use as feedback
                                           during training, if feedback is enabled.
                                           (default: {None})
            init_state {np.ndarray} -- Initialization vector for states.
            (default: {None})
            init_fb {np.ndarray} -- Initialization vector for feedback.
            (default: {None})
            wash_nr_time_step {int} -- Number of states to considered as transitory
                            when training. (default: {0})
            input_id {int} -- Index of the input in the queue. Used for parallelization
                              of computations. (default: {None})

        Raises:
            RuntimeError: raised if no teachers are specified for training
            with feedback.

        Returns:
            Union[Tuple[np.ndarray, np.ndarray], np.ndarray] -- Index of the
            input in queue
            and computed states, or just states if no index is provided.
        """
        if self.Wfb is not None and forced_teacher is None and self.Wout is None:
            raise RuntimeError(
                "Impossible to use feedback without readout "
                "matrix or teacher forcing."
            )

        check_input_lists(input, self.dim_in, forced_teacher, self.dim_out)

        # to track successive internal states of the reservoir
        states = np.zeros((len(input), self.N), dtype=self.typefloat)

        # if a feedback matrix is available, feedback will be set to 0 or to
        # a specific value.
        if init_fb is not None:
            last_feedback = init_fb.copy().reshape(1, -1)
        else:
            last_feedback = self.zero_feedback()

        # State is initialized to 0 or to a specific value.
        if init_state is None:
            current_state = self.zero_state()
        else:
            current_state = init_state.copy().reshape(1, -1)

        # random generator for training/running with additive noise
        rng = default_rng(seed)

        pbar = None
        if kwargs.get("pbar") is not None:
            pbar = kwargs.get("pbar")
        elif verbose is True:
            pbar = tqdm(total=input.shape[0])

        # for each time step in the input
        for t in range(input.shape[0]):
            # compute next state from current state
            current_state = self._get_next_state(
                input[t, :],
                feedback=last_feedback,
                last_state=current_state,
                noise_generator=rng,
            )

            # compute last feedback
            if self.Wfb is not None:
                # during training outputs are equal to teachers for feedback
                if forced_teacher is not None:
                    last_feedback = forced_teacher[t, :]
                # feedback of outputs, computed with Wout
                else:
                    last_feedback = add_bias(current_state) @ self.Wout.T

            states[t, :] = current_state

            if pbar is not None:
                pbar.update(1)

        return states

    def compute_all_states(
        self,
        inputs: Sequence[np.ndarray],
        forced_teachers: Sequence[np.ndarray] = None,
        init_state: np.ndarray = None,
        init_fb: np.ndarray = None,
        workers: int = -1,
        seed: int = None,
        verbose: bool = False,
    ) -> Sequence[np.ndarray]:
        """Compute all states generated from sequences of inputs.

        Parameters
        ----------
            inputs: list or array of numpy.array
                All sequences of inputs used for internal state computation.
                Note that it should always be a list of sequences, i.e. if
                only one sequence of inputs is used, it should be alone in a
                list

            forced_teachers: list or array of numpy.array, optional
                Sequence of ground truths, for computation with feedback without
                any trained readout. Note that is should always be a list of
                sequences of the same length than the `inputs`, i.e. if
                only one sequence of inputs is used, it should be alone in a
                list.

            init_state: np.ndarray, optional
                State initialization vector for all inputs. By default, state
                is initialized at 0.

            init_fb: np.ndarray, optional
                Feedback initialization vector for all inputs, if feedback is
                enabled. By default, feedback is initialized at 0.

            workers: int, optional
                If n >= 1, will enable parallelization of states computation with
                n threads/processes, if possible. If n = -1, will use all available
                resources for parallelization. By default, -1.

            verbose: bool, optional

        Returns
        -------
            list of np.ndarray
                All computed states.
        """
        inputs, forced_teachers = check_input_lists(
            inputs, self.dim_in, forced_teachers, self.dim_out
        )

        workers = min(workers, len(inputs))

        backend = "sequential"
        if workers > 1 or workers == -1:
            backend = get_joblib_backend()

        if forced_teachers is None:
            forced_teachers = [None] * len(inputs)

        compute_states = partial(
            self._compute_states, init_state=init_state, init_fb=init_fb
        )

        steps = np.sum([i.shape[0] for i in inputs])

        progress = ParallelProgressQueue(total=steps, text="States", verbose=verbose)
        with progress as pbar:
            with Parallel(backend=backend, n_jobs=workers) as parallel:
                states = parallel(
                    delayed(partial(compute_states, pbar=pbar, seed=seed))(x, t)
                    for x, t in zip(inputs, forced_teachers)
                )

        return states

    def compute_outputs(
        self, states: Data, verbose: bool = False
    ) -> Sequence[np.ndarray]:
        """Compute all readouts of a given sequence of states,
        when a readout matrix is available (i.e. after training).

        Parameters
        ----------
            states: list of numpy.array
                All sequences of states used for readout.

            verbose: bool, optional

        Raises
        ------
            RuntimeError: no readout matrix Wout is available.
            Consider training model first, or load an existing matrix.

        Returns
        -------
            list of numpy.arrays
                All outputs of readout matrix.
        """
        states, _ = check_input_lists(states, self.N)
        # because all states and readouts will be concatenated,
        # first save the indexes of each inputs states in the concatenated vector.
        if self.Wout is not None:

            if verbose:
                print("Computing outputs...")
                tic = time.time()

            outputs = [None] * len(states)
            for i, s in enumerate(states):
                x = add_bias(s)
                y = (x @ self.Wout.T).astype(self.typefloat)
                outputs[i] = y

            if verbose:
                toc = time.time()
                print(f"Outputs computed! (in {toc - tic}sec)")

            return outputs

        else:
            raise RuntimeError(
                "Impossible to compute outputs: "
                "no readout matrix available. "
                "Train the network first."
            )

    def train(
        self,
        inputs: Data,
        teachers: Data,
        washout: int = 0,
        workers: int = -1,
        return_states: bool = False,
        seed: int = None,
        verbose: bool = False,
    ) -> Sequence[np.ndarray]:
        raise NotImplementedError

    def run(
        self,
        inputs: Data,
        init_state: np.ndarray = None,
        init_fb: np.ndarray = None,
        workers: int = -1,
        return_states=False,
        backend=None,
        seed: int = None,
        verbose: bool = False,
    ) -> Tuple[Sequence[np.ndarray], Sequence[np.ndarray]]:
        """Run the model on a sequence of inputs, and returned the states and
           readouts vectors.

        Parameters
        ----------
            inputs: list of numpy.ndarray
                List of inputs.
                Note that it should always be a list of sequences, i.e. if
                only one sequence (array with rows representing time axis)
                of inputs is used, it should be alone in a list.

            init_state: numpy.ndarray
                State initialization vector for all inputs. By default, internal
                state of the reservoir is initialized to 0.

            init_fb: numpy.ndarray
                Feedback initialization vector for all inputs, if feedback is
                enabled. By default, feedback is initialized to 0.

           workers: int, optional
                If n >= 1, will enable parallelization of states computation with
                n threads/processes, if possible. If n = -1, will use all available
                resources for parallelization. By default, -1.

            return_states: bool, False by default
                If `True`, the function will return all the internal states computed
                during the training. Be warned that this may be too heavy for the
                memory of your computer.

            verbose: bool, optional

            backend:
                kept for compatibility with previous versions.

        Returns
        -------
            list of numpy.ndarray, list of numpy.ndarray
                All outputs computed from readout and all corresponding internal states,
                for all inputs.

        Note
        ----
            If only one input sequence is provided ("continuous time" inputs),
            workers should be 1, because parallelization is impossible. In other
            cases, if using large NumPy arrays during computation (which is often
            the case), prefer using `threading` backend to avoid huge overhead.
            Multiprocess is a good idea only in very specific cases, and this code
            is not (yet) well suited for this.
        """
        # autochecks of inputs and outputs
        inputs, _ = check_input_lists(inputs, self.dim_in)

        lengths = [i.shape[0] for i in inputs]
        steps = sum(lengths)

        if verbose:
            print(f"Running on {len(inputs)} inputs ({steps} steps)")

        def run_fn(*, x, pbar):
            s = self._compute_states(
                x, init_state=init_state, init_fb=init_fb, seed=seed, pbar=pbar
            )

            out = self.compute_outputs([s])[0]

            return out, s

        outputs, states = parallelize(
            self,
            run_fn,
            workers,
            lengths,
            return_states,
            pbar_text="Run",
            verbose=verbose,
            x=inputs,
        )

        return outputs, states

    def generate(
        self,
        nb_timesteps: int,
        warming_inputs: np.ndarray = None,
        init_state: np.ndarray = None,
        init_fb: np.ndarray = None,
        verbose: bool = False,
        init_inputs: np.ndarray = None,
        seed: int = None,
        return_init: bool = None,
    ) -> Tuple[np.ndarray, np.ndarray]:
        """Run the ESN on generative mode.

        After the `warming_inputs` are consumed, new outputs are
        used as inputs for the next nb_timesteps, i.e. the
        ESN is feeding himself with its own outputs.

        Note that this mode can only work if the ESN is trained
        on a regression task. The outputs of the ESN must be
        the same kind of data as its input.

        To train an ESN on generative mode, use the :py:func:`ESN.train`
        method to train the ESN on a regression task (for
        instance, predict the future data point t+1 of a timeseries
        give the data at time t).

        Parameters
        ----------
            nb_timesteps: int
                Number of timesteps of data to generate
                from the initial input.
            warming_inputs: numpy.ndarray
                Input data used to initiate generative mode.
                This data is meant to "seed" the ESN internal
                states with some real information, before it runs
                on its own created outputs.
            init_state: numpy.ndarray, optional:
                State initialization vector for the reservoir.
                By default, internal state of the reservoir is initialized to 0.
            init_fb: numpy.ndarray, optional
                Feedback initialization vector for the reservoir, if feedback is
                enabled. By default, feedback is initialized to 0.
            verbose: bool, optional
            init_inputs: list of numpy.ndarray, optional
                Same as ``warming_inputs``.
                Kept for compatibility with previous version. Deprecated
                since 0.2.2, will be removed soon.
            return_init: bool, optional
                Kept for compatibility with previous version. Deprecated
                since 0.2.2, will be removed soon.

        Returns
        -------
            tuple of numpy.ndarray
                Generated outputs, generated states, warming outputs, warming states

                Generated outputs are the timeseries predicted by the ESN from
                its own predictions over time. Generated states are the
                corresponding internal states.

                Warming outputs are the predictions made by the ESN based on the
                warming inputs passed as parameters. These predictions are prior
                to the generated outputs. Warming states are the corresponding
                internal states. In the case no warming inputs are provided, warming
                outputs and warming states are None.

        """
        if warming_inputs is None and init_state is None and init_inputs is None:
            raise ValueError(
                "at least one of the parameter 'warming_input' "
                "or 'init_state' must not be None. Impossible "
                "to generate from scratch."
            )

        if return_init is not None:
            warnings.warn(
                "Deprecation warning : return_init parameter "
                "is deprecated since 0.2.2 and will be removed."
            )

        # for additive noise in the reservoir
        # 2 separate seeds made from one: one for the warming
        # (if needed), one for the generation
        seed = seed if seed is not None else self.seed
        ss = SeedSequence(seed)
        child_seeds = ss.spawn(2)

        if warming_inputs is not None or init_inputs is not None:
            if init_inputs is not None:
                warnings.warn(
                    "Deprecation warning : init_inputs parameter "
                    "is deprecated since 0.2.2 and will be removed. "
                    "Please use warming_inputs instead."
                )
                warming_inputs = init_inputs

            if verbose:
                print(
                    f"Generating {nb_timesteps} timesteps from "
                    f"{warming_inputs.shape[0]} inputs."
                )
                print("Computing initial states...")

            warming_states = self._compute_states(
                warming_inputs,
                init_state=init_state,
                init_fb=init_fb,
                seed=child_seeds[0],
            )

            # initial state (at beginning of generation)
            s0 = warming_states[-1, :].reshape(1, -1)
            warming_outputs = self.compute_outputs([warming_states])[0]
            # initial input (at beginning of generation)
            u1 = warming_outputs[-1, :].reshape(1, -1)

            if init_fb is not None:
                # initial feedback (at beginning of generation)
                fb0 = warming_outputs[-2, :].reshape(1, -1)
            else:
                fb0 = None

        else:
            warming_outputs, warming_states = None, None
            # time is often first axis but compute_outputs await
            # for time in second axis, so the reshape :
            s0 = init_state.reshape(1, -1)

            if init_fb is not None:
                fb0 = init_fb.reshape(1, -1)
            else:
                fb0 = None

            u1 = self.compute_outputs([s0])[0][-1, :].reshape(1, -1)

        states = np.zeros((nb_timesteps, self.N))
        outputs = np.zeros((nb_timesteps, self.dim_out))

        if verbose:
            track = tqdm
        else:

            def track(x, text):
                return x

        # for additive noise in the reservoir
        rg = default_rng(child_seeds[1])

        for i in track(range(nb_timesteps), "Generate"):
            # from new input u1 and previous state s0
            # compute next state s1 -> s0
            s1 = self._get_next_state(
                single_input=u1, feedback=fb0, last_state=s0, noise_generator=rg
            )

            s0 = s1[-1, :].reshape(1, -1)
            states[i, :] = s0

            if fb0 is not None:
                fb0 = u1.copy()

            # from new state s1 compute next input u2 -> u1
            u1 = self.compute_outputs([s0])[0][-1, :].reshape(1, -1)
            outputs[i, :] = u1

        return outputs, states, warming_outputs, warming_states

    def save(self, directory: str):
        """Save the ESN to disk.

        Parameters
        ----------
            directory: str or Path
                Directory where to save the model.
        """
        _save(self, directory)



"""
============================================
ReservoirPy Nodes (:mod:`reservoirpy.nodes`)
============================================

.. currentmodule:: reservoirpy.nodes

Reservoirs
==========

.. autosummary::
   :toctree: generated/
   :template: autosummary/class.rst

   Reservoir - Recurrent pool of leaky integrator neurons
   NVAR - Non-linear Vector Autoregressive machine (NG-RC)
   IPReservoir - Reservoir with intrinsic plasticity learning rule
   LocalPlasticityReservoir - Reservoir with weight plasticity

Offline readouts
================

.. autosummary::
   :toctree: generated/
   :template: autosummary/class.rst

   Ridge - Layer of neurons connected through offline linear regression.
   ScikitLearnNode - Interface for linear models from the scikit-learn library.

Online readouts
===============

.. autosummary::
   :toctree: generated/
   :template: autosummary/class.rst

   LMS - Layer of neurons connected through least mean squares learning rule.
   RLS - Layer of neurons connected through recursive least squares learning rule.
   FORCE - Layer of neurons connected through online learning rules.

Optimized ESN
=============

.. autosummary::
   :toctree: generated/
   :template: autosummary/class.rst

   ESN - Echo State Network model with distributed offline learning.

Activation functions
====================

.. autosummary::
   :toctree: generated/
   :template: autosummary/class.rst

   Tanh - Hyperbolic tangent node.
   Sigmoid - Logistic function node.
   Softmax - Softmax node.
   Softplus - Softplus node.
   ReLU - Rectified Linear Unit node.
   Identity - Identity function node.

Input and Output
================

.. autosummary::
   :toctree: generated/
   :template: autosummary/class.rst

   Input - Input node, used to distribute input data to other nodes.
   Output - Output node, used to gather stated from hidden nodes.

Operators
=========

.. autosummary::
   :toctree: generated/
   :template: autosummary/class.rst

   Concat - Concatenate vector of data along feature axis.
   Delay - Adds a discrete delay between input and output.
"""
# Author: Nathan Trouvain at 16/12/2021 <nathan.trouvain@inria.fr>
# Licence: MIT License
# Copyright: Xavier Hinaut (2018) <xavier.hinaut@inria.fr>
from .activations import Identity, ReLU, Sigmoid, Softmax, Softplus, Tanh
from .concat import Concat
from .delay import Delay
from .esn import ESN
from .io import Input, Output
from .readouts import FORCE, LMS, RLS, Ridge, ScikitLearnNode
from .reservoirs import NVAR, IPReservoir, LocalPlasticityReservoir, Reservoir

__all__ = [
    "Reservoir",
    "Input",
    "Output",
    "Ridge",
    "FORCE",
    "LMS",
    "RLS",
    "Tanh",
    "Softmax",
    "Softplus",
    "Identity",
    "Sigmoid",
    "ReLU",
    "NVAR",
    "ESN",
    "Concat",
    "Delay",
    "IPReservoir",
    "ScikitLearnNode",
    "LocalPlasticityReservoir",
]



# Author: Nathan Trouvain at 19/06/2021 <nathan.trouvain@inria.fr>
# Licence: MIT License
# Copyright: Xavier Hinaut (2018) <xavier.hinaut@inria.fr>
import gc
import os
import sys
import uuid
from collections import defaultdict

import numpy as np

from ..type import global_dtype

_AVAILABLE_BACKENDS = ("loky", "multiprocessing", "threading", "sequential")

# FIX waiting for a workaround to avoid crashing with multiprocessing
# activated with Python < 3.8. Seems to be due to compatibility issues
# with pickle5 protocol and loky library.
if sys.version_info < (3, 8):
    _BACKEND = "sequential"
else:
    _BACKEND = "loky"

temp_registry = defaultdict(list)


def get_joblib_backend(workers=-1, backend=None):
    if backend is not None:
        if sys.version_info < (3, 8):
            return "sequential"
        if backend in _AVAILABLE_BACKENDS:
            return backend
        else:
            raise ValueError(
                f"'{backend}' is not a Joblib backend. Available "
                f"backends are {_AVAILABLE_BACKENDS}."
            )
    return _BACKEND if workers > 1 or workers == -1 else "sequential"


def set_joblib_backend(backend):
    global _BACKEND
    if backend in _AVAILABLE_BACKENDS:
        _BACKEND = backend
    else:
        raise ValueError(
            f"'{backend}' is not a valid joblib "
            f"backend value. Available backends are "
            f"{_AVAILABLE_BACKENDS}."
        )


def memmap_buffer(node, data=None, shape=None, dtype=None, mode="w+", name=None):

    from .. import _TEMPDIR

    global temp_registry

    caller = node.name
    if data is None:
        if shape is None:
            raise ValueError(
                f"Impossible to create buffer for node {node}: "
                f"neither data nor shape were given."
            )

    temp = os.path.join(_TEMPDIR, f"{caller}-{name}-{uuid.uuid4()}")

    temp_registry[node].append(temp)

    shape = shape if shape is not None else data.shape
    dtype = dtype if dtype is not None else global_dtype

    memmap = np.memmap(temp, shape=shape, mode=mode, dtype=dtype)

    if data is not None:
        memmap[:] = data

    return memmap


def clean_tempfile(caller):
    global temp_registry

    gc.collect()
    for file in temp_registry[caller]:
        try:
            os.remove(file)
        except OSError:
            pass



# Author: Nathan Trouvain at 06/10/2021 <nathan.trouvain@inria.fr>
# Licence: MIT License
# Copyright: Xavier Hinaut (2018) <xavier.hinaut@inria.fr>
import numpy as np

from ..activationsfunc import relu, tanh
from ..node import Node


def forward(node: Node, x: np.ndarray) -> np.ndarray:
    store = node.store
    beta = node.beta

    new_store = np.roll(store, -1)
    new_store[-1] = x

    node.set_param("store", new_store)

    sigma = np.std(new_store)

    if sigma < 1e-8:
        sigma = 1e-8

    x_norm = (x - np.mean(new_store)) / sigma

    return relu(tanh(x_norm / beta))


def initialize(node, x=None, *args, **kwargs):
    if x is not None:
        node.set_input_dim(x.shape[1])
        node.set_output_dim(x.shape[1])

        window = node.window

        node.set_param("store", np.zeros((window, node.output_dim)))


class AsabukiNorm(Node):
    def __init__(self, window, beta=3, name=None):
        super(AsabukiNorm, self).__init__(
            params={"store": None},
            hypers={"window": window, "beta": beta},
            forward=forward,
            initializer=initialize,
            name=name,
        )



# Author: Nathan Trouvain at 18/06/2021 <nathan.trouvain@inria.fr>
# Licence: MIT License
# Copyright: Xavier Hinaut (2018) <xavier.hinaut@inria.fr>
import numpy as np
import pytest

from ..regression_models import RidgeRegression


@pytest.fixture(scope="session")
def dummy_clf_data():
    Xn0 = np.array([[np.sin(x), np.cos(x)] for x in np.linspace(0, 4 * np.pi, 250)])
    Xn1 = np.array(
        [
            [np.sin(10 * x), np.cos(10 * x)]
            for x in np.linspace(np.pi / 4, 4 * np.pi + np.pi / 4, 250)
        ]
    )
    X = np.vstack([Xn0, Xn1])
    y = np.r_[np.zeros(250), np.ones(250)].reshape(-1, 1)

    return X, y


@pytest.fixture(scope="session")
def dummy_data():
    X = np.ones(shape=(200, 50))
    Y = np.ones(shape=(200, 5))

    return X, Y


@pytest.fixture(scope="session")
def bad_xdata():
    X = np.ones(shape=(200, 50))
    Y = np.ones(shape=(200, 5))

    bad_x = np.ones(shape=(199, 50))

    return [X, bad_x, X], [Y, Y, Y]


@pytest.fixture(scope="session")
def bad_ydata():
    X = np.ones(shape=(200, 50))
    Y = np.ones(shape=(200, 5))

    bad_y = np.ones(shape=(200, 4))

    return [X, X, X], [Y, bad_y, Y]


def test_ridge_regression(dummy_data):
    model = RidgeRegression(ridge=0.1)
    model.initialize(50, 5)

    X, Y = dummy_data
    for x, y in zip(X, Y):
        model.partial_fit(x, y)
        XXT = model._XXT.copy()
        YXT = model._YXT.copy()

    assert XXT.shape == (51, 51)
    assert YXT.shape == (5, 51)

    w = model.fit()

    assert w.shape == (5, 51)

    w = model.fit(X, Y)
    assert w.shape == (5, 51)

    for x, y in zip([X, X, X], [Y, Y, Y]):
        model.partial_fit(x, y)
        XXT = model._XXT.copy()
        YXT = model._YXT.copy()

    assert XXT.shape == (51, 51)
    assert YXT.shape == (5, 51)

    w = model.fit()

    assert w.shape == (5, 51)

    w = model.fit([X, X, X], [Y, Y, Y])
    assert w.shape == (5, 51)


def test_ridge_regression_raises(bad_xdata, bad_ydata):
    model = RidgeRegression(ridge=0.1)
    model.initialize(50, 5)

    X, Y = bad_xdata
    with pytest.raises(ValueError):
        for x, y in zip(X, Y):
            model.partial_fit(x, y)
            XXT = model._XXT.copy()
            YXT = model._YXT.copy()

    X, Y = bad_ydata
    with pytest.raises(ValueError):
        for x, y in zip(X, Y):
            model.partial_fit(x, y)
            XXT = model._XXT.copy()
            YXT = model._YXT.copy()



# Author: Nathan Trouvain at 06/10/2021 <nathan.trouvain@inria.fr>
# Licence: MIT License
# Copyright: Xavier Hinaut (2018) <xavier.hinaut@inria.fr>
import numpy as np

from ..node import Node
from ..utils.random import rand_generator


def forward(node, x):
    choice = node.choice
    return x[:, choice.astype(int)]


def initialize(node, x=None, *args, **kwargs):
    if x is not None:
        node.set_input_dim(x.shape[1])
        node.set_output_dim(node.n)

        choice = rand_generator(node.seed).choice(
            np.arange(x.shape[1]), node.n, replace=False
        )

        node.set_param("choice", choice)


class RandomChoice(Node):
    def __init__(self, n, seed=None, name=None):
        super(RandomChoice, self).__init__(
            params={"choice": None},
            hypers={"n": n},
            forward=forward,
            initializer=initialize,
            name=name,
        )
        self.seed = seed



# Author: Nathan Trouvain at 08/07/2021 <nathan.trouvain@inria.fr>
# Licence: MIT License
# Copyright: Xavier Hinaut (2018) <xavier.hinaut@inria.fr>
from typing import Sequence

import numpy as np

from ..node import Node


def concat_forward(concat: Node, data):
    axis = concat.axis

    if not isinstance(data, np.ndarray):
        if len(data) > 1:
            return np.concatenate(data, axis=axis)
        else:
            return np.asarray(data)
    else:
        return data


def concat_initialize(concat: Node, x=None, **kwargs):
    if x is not None:
        if isinstance(x, np.ndarray):
            concat.set_input_dim(x.shape[1])
            concat.set_output_dim(x.shape[1])
        elif isinstance(x, Sequence):
            result = concat_forward(concat, x)
            concat.set_input_dim(tuple([u.shape[1] for u in x]))
            if result.shape[0] > 1:
                concat.set_output_dim(result.shape)
            else:
                concat.set_output_dim(result.shape[1])


class Concat(Node):
    """Concatenate vector of data along feature axis.

    This node is automatically created behind the scene when a node receives the input
    of more than one node.

    For more information on input concatenation, see
    :ref:`/user_guide/advanced_demo.ipynb#Input-to-readout-connections`

    :py:attr:`Concat.hypers` **list**

    ============= ======================================================================
    ``axis``      Concatenation axis.
    ============= ======================================================================

    Examples
    --------

    >>> x1 = np.arange(0., 10.).reshape(10, 1)
    >>> x2 = np.arange(100., 110.).reshape(10, 1)
    >>>
    >>> from reservoirpy.nodes import Concat
    >>> concat_node = Concat()
    >>>
    >>> out = concat_node.run((x1, x2))
    >>> print(out.T)
    [[  0.   1.   2.   3.   4.   5.   6.   7.   8.   9.]
    [100. 101. 102. 103. 104. 105. 106. 107. 108. 109.]]
    >>> print(out.shape)
    (10, 2)
    """

    def __init__(self, axis=1, name=None):
        super(Concat, self).__init__(
            hypers={"axis": axis},
            forward=concat_forward,
            initializer=concat_initialize,
            name=name,
        )



# Author: Nathan Trouvain at 24/09/2021 <nathan.trouvain@inria.fr>
# Licence: MIT License
# Copyright: Xavier Hinaut (2018) <xavier.hinaut@inria.fr>
# Author: Nathan Trouvain at 06/08/2021 <nathan.trouvain@inria.fr>
# Licence: MIT License
# Copyright: Xavier Hinaut (2018) <xavier.hinaut@inria.fr>
import os
import sys

import numpy as np
from joblib import Parallel, delayed
from numpy.testing import assert_array_almost_equal

from reservoirpy.node import _filter_where_na_target
from reservoirpy.nodes import Reservoir, Ridge


def test_ridge_init():

    node = Ridge(10, ridge=1e-8)

    data = np.ones((1, 100))
    res = node(data)

    assert node.Wout.shape == (100, 10)
    assert node.bias.shape == (1, 10)
    assert node.ridge == 1e-8

    data = np.ones((10000, 100))
    res = node.run(data)

    assert res.shape == (10000, 10)


def test_ridge_partial_fit():

    node = Ridge(10, ridge=1e-8)

    X, Y = np.ones((5, 200, 100)), np.ones((5, 200, 10))
    res = node.fit(X, Y)

    assert node.Wout.shape == (100, 10)
    assert_array_almost_equal(node.Wout, np.ones((100, 10)) * 0.01, decimal=4)
    assert node.bias.shape == (1, 10)
    assert_array_almost_equal(node.bias, np.ones((1, 10)) * 0.01, decimal=4)

    node = Ridge(10, ridge=1e-8)

    X, Y = np.ones((5, 200, 100)), np.ones((5, 200, 10))

    for x, y in zip(X, Y):
        res = node.partial_fit(x, y)

    node.fit()

    assert node.Wout.shape == (100, 10)
    assert_array_almost_equal(node.Wout, np.ones((100, 10)) * 0.01, decimal=4)
    assert node.bias.shape == (1, 10)
    assert_array_almost_equal(node.bias, np.ones((1, 10)) * 0.01, decimal=4)

    data = np.ones((100, 100))
    res = node.run(data)

    assert res.shape == (100, 10)


def test_esn():

    readout = Ridge(10, ridge=1e-8)
    reservoir = Reservoir(100)

    esn = reservoir >> readout

    X, Y = np.ones((5, 200, 100)), np.ones((5, 200, 10))
    res = esn.fit(X, Y)

    assert readout.Wout.shape == (100, 10)
    assert readout.bias.shape == (1, 10)

    data = np.ones((100, 100))
    res = esn.run(data)

    assert res.shape == (100, 10)


def test_ridge_feedback():

    readout = Ridge(10, ridge=1e-8)
    reservoir = Reservoir(100)

    esn = reservoir >> readout

    reservoir <<= readout

    X, Y = np.ones((5, 200, 100)), np.ones((5, 200, 10))
    res = esn.fit(X, Y)

    assert readout.Wout.shape == (100, 10)
    assert readout.bias.shape == (1, 10)
    assert reservoir.Wfb.shape == (100, 10)

    data = np.ones((100, 100))
    res = esn.run(data)

    assert res.shape == (100, 10)


def test_hierarchical_esn():

    reservoir1 = Reservoir(100, input_dim=5, name="h1")
    readout1 = Ridge(ridge=1e-8, name="r1")

    reservoir2 = Reservoir(100, name="h2")
    readout2 = Ridge(ridge=1e-8, name="r2")

    esn = reservoir1 >> readout1 >> reservoir2 >> readout2

    X, Y = np.ones((1, 200, 5)), {
        "r1": np.ones((1, 200, 10)),
        "r2": np.ones((1, 200, 3)),
    }

    res = esn.fit(X, Y)

    assert readout1.Wout.shape == (100, 10)
    assert readout1.bias.shape == (1, 10)

    assert readout2.Wout.shape == (100, 3)
    assert readout2.bias.shape == (1, 3)

    assert reservoir1.Win.shape == (100, 5)
    assert reservoir2.Win.shape == (100, 10)

    data = np.ones((100, 5))
    res = esn.run(data)

    assert res.shape == (100, 3)


def test_parallel():
    if sys.platform in ["win32", "cygwin"] and sys.version_info < (3, 8):
        # joblib>=1.3.0 & Windows & Python<3.8 & loky combined are incompatible
        # see https://github.com/joblib/loky/issues/411
        return

    process_count = 4 * os.cpu_count()

    rng = np.random.default_rng(seed=42)
    x = rng.random((40000, 10))
    y = x[:, 2::-1] + rng.random((40000, 3)) / 10
    x_run = rng.random((20, 10))

    def run_ridge(i):
        readout = Ridge(ridge=1e-8)
        return readout.fit(x, y).run(x_run)

    parallel = Parallel(n_jobs=process_count, return_as="generator")
    results = list(parallel(delayed(run_ridge)(i) for i in range(process_count)))

    for result in results:
        assert np.all(result == results[0])


def test_na():
    # Define X, Y, Y_na
    X = np.ones((10, 5))
    Y = np.ones((10, 5))
    Y_na = Y
    Y_na[-1] = np.nan

    readout = Ridge(ridge=1e-7)

    # Filtered the NaN data and get the predictions | Last row removed from X and Y
    X_filtered, Y_filtered = _filter_where_na_target(X, Y_na)
    readout.fit(X_filtered, Y_filtered)
    pred_filtered = readout.run(X_filtered)

    # Manually deleted the last row of the classic inputs
    readout.fit(X[:-1, :], Y[:-1, :])
    pred_classic = readout.run(X[:-1, :])

    # Assert that those predictions are equal
    assert_array_almost_equal(pred_filtered, pred_classic)



# Author: Nathan Trouvain at 18/11/2021 <nathan.trouvain@inria.fr>
# Licence: MIT License
# Copyright: Xavier Hinaut (2018) <xavier.hinaut@inria.fr>
import numpy as np
from numpy.testing import assert_array_equal

from ..concat import Concat
from ..reservoirs import Reservoir


def test_concat():
    x = [np.ones((1, 5)) for _ in range(3)]

    c = Concat()

    res = c(x)

    assert c.input_dim == (5, 5, 5)
    assert_array_equal(res, np.ones((1, 15)))


def test_concat_no_list():
    x = np.ones((1, 5))

    c = Concat()

    res = c(x)

    assert_array_equal(res, np.ones((1, 5)))


def test_reservoir_union():

    reservoirs = [Reservoir(10, name=f"r{i}") for i in range(3)]

    model = reservoirs >> Concat()

    x = {f"r{i}": np.ones((1, 5)) for i in range(3)}

    res = model(x)

    assert res.shape == (1, 30)

    res_final = Reservoir(20)

    model = reservoirs >> res_final

    res = model(x)

    assert res.shape == (1, 20)
    assert any([isinstance(n, Concat) for n in model.nodes])



from functools import partial
from typing import Callable, Dict, Optional, Sequence, Union

import numpy as np
import scipy.sparse as sp

from ..._base import check_xy
from ...activationsfunc import get_function, identity, tanh
from ...mat_gen import bernoulli, uniform
from ...node import Unsupervised, _init_with_sequences
from ...type import Weights
from ...utils.random import noise, rand_generator
from ...utils.validation import is_array
from .base import forward_external
from .base import initialize as initialize_base
from .base import initialize_feedback


def local_synaptic_plasticity(reservoir, pre_state, post_state):
    """
    Apply the local learning rule (Oja, Anti-Oja, Hebbian, Anti-Hebbian, BCM)
    to update the recurrent weight matrix W.

    If `synapse_normalization=True`, then each row of W is L2-normalized
    immediately after the local rule update.

    This version supports both dense and sparse matrices. For sparse matrices,
    the weight matrix is converted to LIL format for efficient row modifications.
    """

    W = reservoir.W  # Expecting W to be in CSR format.
    increment = reservoir.increment
    do_norm = reservoir.synapse_normalization

    # Extract presynaptic and postsynaptic vectors.
    x = pre_state[0]  # shape: (units,)
    y = post_state[0]  # shape: (units,)

    # Ensure W is in CSR format.
    if not sp.isspmatrix_csr(W):
        W = sp.csr_matrix(W)

    # Compute the row index for each nonzero element using np.repeat.
    # np.diff(W.indptr) gives the count of nonzeros in each row.
    rows = np.repeat(np.arange(W.shape[0]), np.diff(W.indptr))
    cols = W.indices
    data = W.data  # Update in place.

    # Vectorized update of nonzero elements based on the chosen rule.
    data += increment(data, x[cols], y[rows])

    # Optionally normalize each row.
    if do_norm:
        # Compute the L2 norm per row for the updated data.
        row_sums = np.bincount(rows, weights=data**2, minlength=W.shape[0])
        row_norms = np.sqrt(row_sums)
        safe_norms = np.where(row_norms > 0, row_norms, 1)
        data /= safe_norms[rows]

    return W


def sp_backward(reservoir, X=None, *args, **kwargs):
    """
    Offline learning method for the local-rule-based reservoir.
    """
    for epoch in range(reservoir.epochs):
        for seq in X:
            for u in seq:
                pre_state = reservoir.internal_state  # shape (1, units)
                post_state = reservoir.call(u.reshape(1, -1))  # shape (1, units)

                # Update W with the chosen local rule
                W_new = local_synaptic_plasticity(reservoir, pre_state, post_state)
                reservoir.set_param("W", W_new)


def initialize_synaptic_plasticity(reservoir, *args, **kwargs):
    """
    Custom initializer for the LocalRuleReservoir.
    Reuses the ESN-like initialization and sets the reservoir internal state to zeros.
    """

    initialize_base(reservoir, *args, **kwargs)


class LocalPlasticityReservoir(Unsupervised):
    """
    A reservoir that learns its recurrent weights W through a local
    learning rule selected by the 'learning_rule' hyperparameter.

    Reservoir states are updated with the external equation:

    .. math::

        r[t+1] = (1 - lr)*r[t] + lr*(W r[t] + Win u[t+1] + Wfb fb[t] + bias)
        x[t+1] = activation(r[t+1])

    Then the local rule is applied each timestep to update W.

    .. math::

        W_{ij} \\leftarrow W_{ij} + \\Delta W_{ij}

    Supported rules:
      `oja`:
        :math:`\\Delta W_{ij} = \\eta y (x - y W_{ij})`
      `anti-oja` [1]_ [2]_ [3]_ :
        :math:`\\Delta W_{ij} = - \\eta y (x - y W_{ij})`
      `hebbian` [4]_ :
        :math:`\\Delta W_{ij} = \\eta x y`
      `anti-hebbian`:
        :math:`\\Delta W_{ij} = - \\eta x y`
      `bcm` [2]_ :
        :math:`\\Delta W_{ij} = \\eta x y (y - \\theta_{BCM})`

    Where :math:`x` represents the pre-synaptic state and :math:`y` represents
    the post-synaptic state of the neuron.

    For "bcm", you can set a threshold 'bcm_theta' (default `0.0`).

    If `synapse_normalization=True`, then after each local-rule update
    on a row i of W, the row is rescaled to unit L2 norm. [4]_



    Parameters
    ----------
    units : int, optional
        Number of reservoir units. If None, the number of units will be inferred from
        the ``W`` matrix shape.
    local_rule : str, default to `oja`
        One of `"oja"`, `"anti-oja"`, `"hebbian"`, `"anti-hebbian"`, `"bcm"`.
    eta : float, default to 1e-3.
        Local learning rate for the weight update.
    bcm_theta : float, default to 0.0
        The threshold used in the "bcm" rule.
    synapse_normalization : bool, default to True
        If True, L2-normalize each row of W after its update.
    epochs : int, default to 1
        Number of training iterations.
    sr : float, optional
        Spectral radius of recurrent weight matrix.
    lr : float or array-like of shape (units,), default to 1.0
        Neurons leak rate. Must be in :math:`[0, 1]`.
    input_bias : bool, default to True
        If False, no bias is added to inputs.
    noise_rc : float, default to 0.0
        Gain of noise applied to reservoir activations.
    noise_in : float, default to 0.0
        Gain of noise applied to input inputs.
    noise_fb : float, default to 0.0
        Gain of noise applied to feedback signal.
    noise_type : str, default to "normal"
        Distribution of noise. Must be a Numpy random variable generator
        distribution (see :py:class:`numpy.random.Generator`).
    noise_kwargs : dict, optional
        Keyword arguments to pass to the noise generator, such as `low` and `high`
        values of uniform distribution.
    input_scaling : float or array-like of shape (features,), default to 1.0.
        Input gain. An array of the same dimension as the inputs can be used to
        set up different input scaling for each feature.
    bias_scaling: float, default to 1.0
        Bias gain.
    fb_scaling : float or array-like of shape (features,), default to 1.0
        Feedback gain. An array of the same dimension as the feedback can be used to
        set up different feedback scaling for each feature.
    input_connectivity : float, default to 0.1
        Connectivity of input neurons, i.e. ratio of input neurons connected
        to reservoir neurons. Must be in :math:`]0, 1]`.
    rc_connectivity : float, default to 0.1
        Connectivity of recurrent weight matrix, i.e. ratio of reservoir
        neurons connected to other reservoir neurons, including themselves.
        Must be in :math:`]0, 1]`.
    fb_connectivity : float, default to 0.1
        Connectivity of feedback neurons, i.e. ratio of feedback neurons
        connected to reservoir neurons. Must be in :math:`]0, 1]`.
    Win : callable or array-like of shape (units, features), default to :py:func:`~reservoirpy.mat_gen.bernoulli`
        Input weights matrix or initializer. If a callable (like a function) is
        used,
        then this function should accept any keywords
        parameters and at least two parameters that will be used to define the
        shape of
        the returned weight matrix.
    W : callable or array-like of shape (units, units), default to :py:func:`~reservoirpy.mat_gen.uniform`
        Recurrent weights matrix or initializer. If a callable (like a function) is
        used, then this function should accept any keywords
        parameters and at least two parameters that will be used to define the
        shape of
        the returned weight matrix.
    Wfb : callable or array-like of shape (units, feedback), default to :py:func:`~reservoirpy.mat_gen.bernoulli`
        Feedback weights matrix or initializer. If a callable (like a function) is
        used, then this function should accept any keywords
        parameters and at least two parameters that will be used to define the
        shape of
        the returned weight matrix.
    bias : callable or array-like of shape (units, 1), default to :py:func:`~reservoirpy.mat_gen.bernoulli`
        Bias weights vector or initializer. If a callable (like a function) is
        used, then this function should accept any keywords
        parameters and at least two parameters that will be used to define the
        shape of
        the returned weight matrix.
    feedback_dim : int, optional
        Feedback dimension. Can be inferred at first call.
    fb_activation : str or callable, default to :py:func:`~reservoirpy.activationsfunc.identity`
        Feedback activation function.
        - If a str, should be a :py:mod:`~reservoirpy.activationsfunc`
        function name.
        - If a callable, should be an element-wise operator on arrays.
    activation : {"tanh", "sigmoid"}, default to "tanh"
        Reservoir units activation function.
    input_dim : int, optional
        Input dimension. Can be inferred at first call.
    name : str, optional
        Node name.
    dtype : Numpy dtype, default to np.float64
        Numerical type for node parameters.
    seed : int or :py:class:`numpy.random.Generator`, optional
        A random state seed, for noise generation.


    References
    ----------

    .. [1] Babinec, Š., & Pospíchal, J. (2007). Improving the prediction
           accuracy of echo state neural networks by anti-Oja’s learning.
           In International Conference on Artificial Neural Networks (pp. 19-28).
           Berlin, Heidelberg: Springer Berlin Heidelberg.
           https://doi.org/10.1007/978-3-540-74690-4_3

    .. [2] Yusoff, M. H., Chrol-Cannon, J., & Jin, Y. (2016). Modeling neural
           plasticity in echo state networks for classification and regression.
           Information Sciences, 364, 184-196.
           https://doi.org/10.1016/j.ins.2015.11.017

    .. [3] Morales, G. B., Mirasso, C. R., & Soriano, M. C. (2021). Unveiling
           the role of plasticity rules in reservoir computing. Neurocomputing,
           461, 705-715. https://doi.org/10.1016/j.neucom.2020.05.127

    .. [4] Wang, X., Jin, Y., & Hao, K. (2021). Synergies between synaptic and
           intrinsic plasticity in echo state networks. Neurocomputing,
           432, 32-43. https://doi.org/10.1016/j.neucom.2020.12.007

    Example
    -------
    >>> reservoir = LocalPlasticityReservoir(
    ...     units=100, sr=0.9, local_rule="hebbian",
    ...     eta=1e-3, epochs=5, synapse_normalization=True
    ... )
    >>> # Fit on data timeseries
    >>> reservoir.fit(X_data, warmup=10)
    >>> # Then run
    >>> states = reservoir.run(X_data)
    """

    def __init__(
        self,
        units: Optional[int] = None,
        # local rule choice
        local_rule: str = "oja",
        eta: float = 1e-3,
        bcm_theta: float = 0.0,
        synapse_normalization: bool = False,
        epochs: int = 1,
        # standard reservoir params
        sr: Optional[float] = None,
        lr: float = 1.0,
        input_bias: bool = True,
        noise_rc: float = 0.0,
        noise_in: float = 0.0,
        noise_fb: float = 0.0,
        noise_type: str = "normal",
        noise_kwargs: Optional[Dict] = None,
        input_scaling: Union[float, Sequence] = 1.0,
        bias_scaling: float = 1.0,
        fb_scaling: Union[float, Sequence] = 1.0,
        input_connectivity: Optional[float] = 0.1,
        rc_connectivity: Optional[float] = 0.1,
        fb_connectivity: Optional[float] = 0.1,
        Win: Union[Weights, Callable] = bernoulli,
        W: Union[Weights, Callable] = uniform,
        Wfb: Union[Weights, Callable] = bernoulli,
        bias: Union[Weights, Callable] = bernoulli,
        feedback_dim: Optional[int] = None,
        fb_activation: Union[str, Callable] = identity,
        activation: Union[str, Callable] = tanh,
        name=None,
        seed=None,
        **kwargs,
    ):
        if units is None and not is_array(W):
            raise ValueError(
                "'units' parameter must not be None if 'W' parameter is not a matrix."
            )

        rng = rand_generator(seed=seed)
        noise_kwargs = dict() if noise_kwargs is None else noise_kwargs

        # Validate local rule name
        local_rule = local_rule.lower()

        if local_rule == "oja":

            def increment(weights, pre_state, post_state):
                return eta * post_state * (pre_state - post_state * weights)

        elif local_rule == "anti-oja":

            def increment(weights, pre_state, post_state):
                return -eta * post_state * (pre_state - post_state * weights)

        elif local_rule == "hebbian":

            def increment(weights, pre_state, post_state):
                return eta * post_state * pre_state

        elif local_rule == "anti-hebbian":

            def increment(weights, pre_state, post_state):
                return -eta * post_state * pre_state

        elif local_rule == "bcm":

            def increment(weights, pre_state, post_state):
                return eta * post_state * (post_state - bcm_theta) * pre_state

        else:
            raise ValueError(
                f"Unknown learning rule '{local_rule}'. Choose from: "
                "['oja', 'anti-oja', 'hebbian', 'anti-hebbian', 'bcm']."
            )

        self.increment = increment

        super(LocalPlasticityReservoir, self).__init__(
            fb_initializer=partial(
                initialize_feedback,
                Wfb_init=Wfb,
                fb_scaling=fb_scaling,
                fb_connectivity=fb_connectivity,
                seed=seed,
            ),
            params={
                "W": None,
                "Win": None,
                "Wfb": None,
                "bias": None,
                "internal_state": None,
            },
            hypers={
                "local_rule": local_rule,
                "bcm_theta": bcm_theta,
                "eta": eta,
                "synapse_normalization": synapse_normalization,
                "sr": sr,
                "lr": lr,
                "epochs": epochs,
                "input_bias": input_bias,
                "input_scaling": input_scaling,
                "fb_scaling": fb_scaling,
                "rc_connectivity": rc_connectivity,
                "input_connectivity": input_connectivity,
                "fb_connectivity": fb_connectivity,
                "noise_in": noise_in,
                "noise_rc": noise_rc,
                "noise_out": noise_fb,
                "noise_type": noise_type,
                "activation": get_function(activation)
                if isinstance(activation, str)
                else activation,
                "fb_activation": get_function(fb_activation)
                if isinstance(fb_activation, str)
                else fb_activation,
                "units": units,
                "noise_generator": partial(noise, rng=rng, **noise_kwargs),
            },
            forward=forward_external,
            initializer=partial(
                initialize_synaptic_plasticity,
                input_bias=input_bias,
                bias_scaling=bias_scaling,
                sr=sr,
                input_scaling=input_scaling,
                input_connectivity=input_connectivity,
                rc_connectivity=rc_connectivity,
                W_init=W,
                Win_init=Win,
                bias_init=bias,
                seed=seed,
            ),
            backward=sp_backward,
            output_dim=units,
            feedback_dim=feedback_dim,
            name=name,
            **kwargs,
        )

    @property
    def fitted(self) -> bool:
        # For an unsupervised node that can always be updated,
        # we set `fitted = True` after first initialization/training.
        return True

    def partial_fit(
        self, X_batch, Y_batch=None, warmup=0, **kwargs
    ) -> "LocalPlasticityReservoir":
        """Partial offline fitting method (for batch training)."""
        X, _ = check_xy(self, X_batch, allow_n_inputs=False)
        X, _ = _init_with_sequences(self, X)

        self.initialize_buffers()

        for i in range(len(X)):
            X_seq = X[i]

            if X_seq.shape[0] <= warmup:
                raise ValueError(
                    f"Warmup set to {warmup} timesteps, "
                    f"but one timeseries is only {X_seq.shape[0]} long."
                )

            # Run warmup if specified
            if warmup > 0:
                self.run(X_seq[:warmup])

            self._partial_backward(self, X_seq[warmup:])

        return self



from math import cos, sin
from tempfile import TemporaryDirectory

import numpy as np
import pytest
from scipy import sparse

from .._esn import ESN
from ..utils.save import load


@pytest.fixture(scope="session")
def matrices():
    Win = np.array([[1, -1], [-1, 1], [1, -1], [-1, -1]])
    W = np.array(
        [
            [0.0, 0.1, -0.1, 0.0],
            [0.2, 0.0, 0.0, -0.2],
            [0.0, 0.2, 0.3, 0.1],
            [-0.1, 0.0, 0.0, 0.0],
        ]
    )

    return W, Win


@pytest.fixture(scope="session")
def matrices_fb():
    Win = np.array([[1, -1], [-1, 1], [1, -1], [-1, -1]])
    W = np.array(
        [
            [0.0, 0.1, -0.1, 0.0],
            [0.2, 0.0, 0.0, -0.2],
            [0.0, 0.2, 0.3, 0.1],
            [-0.1, 0.0, 0.0, 0.0],
        ]
    )
    Wfb = np.array([[1, -1], [-1, -1], [1, 1], [-1, 1]])
    return W, Win, Wfb


@pytest.fixture(scope="session")
def dummy_data():
    Xn0 = np.array([[sin(x), cos(x)] for x in np.linspace(0, 4 * np.pi, 500)])
    Xn1 = np.array(
        [[sin(x), cos(x)] for x in np.linspace(np.pi / 4, 4 * np.pi + np.pi / 4, 500)]
    )
    return Xn0, Xn1


@pytest.fixture(scope="session")
def dummy_clf_data():
    Xn0 = np.array([[sin(x), cos(x)] for x in np.linspace(0, 4 * np.pi, 250)])
    Xn1 = np.array(
        [
            [sin(10 * x), cos(10 * x)]
            for x in np.linspace(np.pi / 4, 4 * np.pi + np.pi / 4, 250)
        ]
    )
    X = np.vstack([Xn0, Xn1])
    y = np.r_[np.zeros(250), np.ones(250)].reshape(-1, 1)

    return X, y


def test_esn(matrices, dummy_data):
    W, Win = matrices
    esn = ESN(lr=0.1, W=W, Win=Win, input_bias=False)
    X, y = dummy_data
    states = esn.train([X], [y], return_states=True)

    assert esn.Wout.shape == (2, 5)

    outputs, states = esn.run([X], return_states=True)

    assert states[0].shape[0] == X.shape[0]
    assert outputs[0].shape[1] == y.shape[1]
    assert outputs[0].shape[0] == states[0].shape[0]

    esn.train([X, X, X], [y, y, y])

    assert esn.Wout.shape == (2, 5)

    outputs, states = esn.run([X, X], return_states=True)

    assert len(states) == 2
    assert len(outputs) == 2


def test_esn_compute_all_states(matrices, dummy_data):
    W, Win = matrices
    esn = ESN(lr=0.1, W=W, Win=Win, input_bias=False)
    X, y = dummy_data
    states = esn.compute_all_states([X])

    assert states[0].shape[0] == X.shape[0]
    assert states[0].shape[1] == esn.N

    states = esn.compute_all_states([X, X, X])

    assert len(states) == 3


def test_esn_generate(matrices, dummy_data):
    W, Win = matrices
    esn = ESN(lr=0.1, W=W, Win=Win, input_bias=False)

    X, y = dummy_data
    states = esn.train([X], [y])
    out, states, warm_out, warm_states = esn.generate(500, warming_inputs=X)

    assert states.shape[0] == 500
    assert out.shape[0] == 500
    assert warm_states.shape[0] == X.shape[0]
    assert warm_out.shape[0] == X.shape[0]


def test_esn_ridge(matrices, dummy_data):
    W, Win = matrices
    esn = ESN(lr=0.1, W=W, Win=Win, input_bias=False, ridge=1e-4)

    X, y = dummy_data
    states = esn.train([X], [y])

    assert esn.Wout.shape == (2, 5)

    outputs, states = esn.run([X], return_states=True)

    assert states[0].shape[0] == X.shape[0]
    assert outputs[0].shape[1] == 2

    states = esn.train([X, X, X], [y, y, y])

    assert esn.Wout.shape == (2, 5)

    outputs, states = esn.run([X, X], return_states=True)

    assert len(states) == 2
    assert len(outputs) == 2


def test_esn_fb(matrices_fb, dummy_data):
    W, Win, Wfb = matrices_fb

    esn = ESN(lr=0.1, W=W, Win=Win, Wfb=Wfb, input_bias=False, fbfunc=np.tanh)
    X, y = dummy_data
    states = esn.train([X], [y])

    assert esn.Wout.shape == (2, 5)

    outputs, states = esn.run([X], return_states=True)

    assert states[0].shape[0] == X.shape[0]
    assert outputs[0].shape[1] == y.shape[1]

    states = esn.train([X, X, X], [y, y, y])

    assert esn.Wout.shape == (2, 5)

    outputs, states = esn.run([X, X], return_states=True)

    assert len(states) == 2
    assert len(outputs) == 2


def test_esn_compute_all_states_fb(matrices_fb, dummy_data):
    W, Win, Wfb = matrices_fb
    esn = ESN(lr=0.1, W=W, Win=Win, Wfb=Wfb, input_bias=False, fbfunc=lambda x: x)
    X, y = dummy_data
    states = esn.compute_all_states([X], [y])

    assert states[0].shape[0] == X.shape[0]
    assert states[0].shape[1] == W.shape[0]

    states = esn.compute_all_states([X, X, X], [y, y, y])

    assert len(states) == 3


def test_esn_compute_all_states_fb_with_readout(matrices_fb, dummy_data):
    W, Win, Wfb = matrices_fb
    esn = ESN(lr=0.1, W=W, Win=Win, Wfb=Wfb, input_bias=False, fbfunc=lambda x: x)
    X, y = dummy_data
    esn.train([X], [y])
    states = esn.compute_all_states([X])

    assert states[0].shape[0] == X.shape[0]
    assert states[0].shape[1] == W.shape[0]

    states = esn.compute_all_states([X, X, X])

    assert len(states) == 3


def test_esn_generate_fb(matrices_fb, dummy_data):
    W, Win, Wfb = matrices_fb
    esn = ESN(lr=0.1, W=W, Win=Win, Wfb=Wfb, input_bias=False, fbfunc=lambda x: x)

    X, y = dummy_data
    states = esn.train([X], [y])
    out, states, warm_out, warm_states = esn.generate(
        500, warming_inputs=X, init_fb=y[0]
    )

    assert states.shape[0] == 500
    assert out.shape[0] == 500
    assert warm_states.shape[0] == X.shape[0]
    assert warm_out.shape[0] == X.shape[0]


def test_save(matrices_fb, dummy_data):
    W, Win, Wfb = matrices_fb
    esn = ESN(
        lr=0.1, W=W, Win=Win, Wfb=Wfb, input_bias=False, fbfunc=np.tanh, noise_rc=0.01
    )

    with TemporaryDirectory() as tempdir:

        esn.save(f"{tempdir}/esn")
        esn = load(f"{tempdir}/esn")

        X, y = dummy_data
        states = esn.train([X], [y])

        assert esn.Wout.shape == (2, 5)

        outputs, states = esn.run([X], return_states=True)

        assert states[0].shape[0] == X.shape[0]
        assert outputs[0].shape[1] == y.shape[1]

        esn.save(f"{tempdir}/esn_trained")
        esn = load(f"{tempdir}/esn_trained")

        outputs, states = esn.run([X, X], return_states=True)

        assert len(states) == 2
        assert len(outputs) == 2


def test_save_sparse(matrices_fb, dummy_data):
    W, Win, Wfb = matrices_fb
    W = sparse.csr_matrix(W)
    esn = ESN(
        lr=0.1, W=W, Win=Win, Wfb=Wfb, input_bias=False, fbfunc=np.tanh, noise_rc=0.01
    )

    with TemporaryDirectory() as tempdir:

        esn.save(f"{tempdir}/esn")
        esn = load(f"{tempdir}/esn")

        X, y = dummy_data
        states = esn.train([X], [y])

        assert esn.Wout.shape == (2, 5)

        outputs, states = esn.run([X], return_states=True)

        assert states[0].shape[0] == X.shape[0]
        assert outputs[0].shape[1] == y.shape[1]

        esn.save(f"{tempdir}/esn_trained")
        esn = load(f"{tempdir}/esn_trained")

        assert sparse.issparse(esn.W)

        outputs, states = esn.run([X, X], return_states=True)

        assert len(states) == 2
        assert len(outputs) == 2



# Author: Nathan Trouvain at 18/11/2021 <nathan.trouvain@inria.fr>
# Licence: MIT License
# Copyright: Xavier Hinaut (2018) <xavier.hinaut@inria.fr>
import sys

if sys.version_info < (3, 8):
    from scipy.special import comb
else:
    from math import comb

import numpy as np

from reservoirpy.nodes import NVAR


def _get_output_dim(input_dim, delay, order):
    linear_dim = delay * input_dim
    nonlinear_dim = comb(linear_dim + order - 1, order)
    return int(linear_dim + nonlinear_dim)


def test_nvar_init():

    node = NVAR(3, 2)

    data = np.ones((1, 10))
    res = node(data)

    assert node.store is not None
    assert node.strides == 1
    assert node.delay == 3
    assert node.order == 2

    data = np.ones((10000, 10))
    res = node.run(data)

    assert res.shape == (10000, _get_output_dim(10, 3, 2))


def test_nvar_chain():

    node1 = NVAR(3, 1)
    node2 = NVAR(3, 2, strides=2)

    data = np.ones((1, 10))
    res = (node1 >> node2)(data)

    assert res.shape == (1, _get_output_dim(_get_output_dim(10, 3, 1), 3, 2))



# Author: Nathan Trouvain at 25/03/2022 <nathan.trouvain@inria.fr>
# Licence: MIT License
# Copyright: Xavier Hinaut (2018) <xavier.hinaut@inria.fr>
from collections import defaultdict

import pytest
from tqdm import tqdm

from reservoirpy.utils import (
    _obj_from_kwargs,
    progress,
    safe_defaultdict_copy,
    verbosity,
)


def test_verbosity():
    v = verbosity()
    from reservoirpy.utils import VERBOSITY

    assert v == VERBOSITY
    verbosity(0)
    from reservoirpy.utils import VERBOSITY

    assert VERBOSITY == 0


def test_progress():
    verbosity(0)
    a = [1, 2, 3]
    it = progress(a)
    assert id(it) == id(a)

    verbosity(1)
    it = progress(a)
    assert isinstance(it, tqdm)


def test_defaultdict_copy():

    a = defaultdict(list)

    a["a"].extend([1, 2, 3])
    a["b"] = 2

    b = safe_defaultdict_copy(a)

    assert list(b.values()) == [
        [1, 2, 3],
        [
            2,
        ],
    ]
    assert list(b.keys()) == ["a", "b"]

    a = dict()

    a["a"] = [1, 2, 3]
    a["b"] = 2

    b = safe_defaultdict_copy(a)

    assert list(b.values()) == [
        [1, 2, 3],
        [
            2,
        ],
    ]
    assert list(b.keys()) == ["a", "b"]


def test_obj_from_kwargs():
    class A:
        def __init__(self, a=0, b=2):
            self.a = a
            self.b = b

    a = _obj_from_kwargs(A, {"a": 1})
    assert a.a == 1
    assert a.b == 2



# Author: Nathan Trouvain at 19/06/2021 <nathan.trouvain@inria.fr>
# Licence: MIT License
# Copyright: Xavier Hinaut (2018) <xavier.hinaut@inria.fr>
from collections import defaultdict
from inspect import signature
from typing import Iterable

from tqdm import tqdm

VERBOSITY = 1


def verbosity(level=None):
    global VERBOSITY
    if isinstance(level, int) and VERBOSITY != level:
        VERBOSITY = level
    return VERBOSITY


def progress(it, *args, **kwargs):
    if VERBOSITY > 0:
        return tqdm(it, *args, **kwargs)
    else:
        return it


def safe_defaultdict_copy(d):
    new_d = defaultdict(list)
    for key, item in d.items():
        if isinstance(item, Iterable):
            new_d[key] = list(item)
        else:
            new_d[key] += [item]
    return new_d


def _obj_from_kwargs(klas, kwargs):
    sig = signature(klas.__init__)
    params = list(sig.parameters.keys())
    klas_kwargs = {n: v for n, v in kwargs.items() if n in params}
    return klas(**klas_kwargs)
